@@ -61,6 +61,8 @@ declare FORK=0
 declare OOB=0
 declare C_ARGS=""
 declare S_ARGS=""
+declare PIN_CORE=""
+declare PROVIDER_TESTS=0
 
 declare cur_excludes=""
 declare file_excludes=""
@@ -109,7 +111,8 @@ functional_tests=(
 	"fi_poll -t queue"
 	"fi_poll -t counter"
 	"fi_rdm"
-	"fi_rdm_rma_simple"
+	"fi_rdm -U"
+	"fi_rdm_rma_event"
 	"fi_rdm_rma_trigger"
 	"fi_shared_ctx"
 	"fi_shared_ctx --no-tx-shared-ctx"
@@ -125,19 +128,20 @@ functional_tests=(
 	"fi_rdm_shared_av"
 	"fi_multi_mr -e msg -V"
 	"fi_multi_mr -e rdm -V"
+	"fi_multi_ep -e msg -v"
+	"fi_multi_ep -e rdm -v"
 	"fi_recv_cancel -e rdm -V"
-	"fi_unexpected_msg -e msg -i 10"
-	"fi_unexpected_msg -e rdm -i 10"
-	"fi_unexpected_msg -e msg -S -i 10"
-	"fi_unexpected_msg -e rdm -S -i 10"
-	"fi_inj_complete -e msg"
-	"fi_inj_complete -e rdm"
-	"fi_inj_complete -e dgram"
-	"fi_inj_complete -e msg -SR"
-	"fi_inj_complete -e rdm -SR"
-	"fi_inj_complete -e dgram -SR"
+	"fi_unexpected_msg -e msg -I 10"
+	"fi_unexpected_msg -e rdm -I 10"
+	"fi_msg_inject -A inject -v"
+	"fi_msg_inject -N -A inject -v"
+	"fi_msg_inject -A inj_complete -v"
+	"fi_msg_inject -N -A inj_complete -v"
 	"fi_bw -e rdm -v -T 1"
+	"fi_bw -e rdm -v -T 1 -U"
 	"fi_bw -e msg -v -T 1"
+	"fi_rdm_multi_client -C 10 -I 5"
+	"fi_rdm_multi_client -C 10 -I 5 -U"
 )
 
 short_tests=(
@@ -149,18 +153,28 @@ short_tests=(
 	"fi_rma_bw -e msg -o read -I 5"
 	"fi_rma_bw -e msg -o writedata -I 5"
 	"fi_rma_bw -e rdm -o write -I 5"
+	"fi_rma_bw -e rdm -o write -I 5 -U"
 	"fi_rma_bw -e rdm -o read -I 5"
+	"fi_rma_bw -e rdm -o read -I 5 -U"
 	"fi_rma_bw -e rdm -o writedata -I 5"
+	"fi_rma_bw -e rdm -o writedata -I 5 -U"
 	"fi_rdm_atomic -I 5 -o all"
+	"fi_rdm_atomic -I 5 -o all -U"
 	"fi_rdm_cntr_pingpong -I 5"
 	"fi_multi_recv -e rdm -I 5"
 	"fi_multi_recv -e msg -I 5"
 	"fi_rdm_pingpong -I 5"
+	"fi_rdm_pingpong -I 5 -U"
 	"fi_rdm_pingpong -I 5 -v"
+	"fi_rdm_pingpong -I 5 -v -U"
 	"fi_rdm_tagged_pingpong -I 5"
+	"fi_rdm_tagged_pingpong -I 5 -U"
 	"fi_rdm_tagged_pingpong -I 5 -v"
+	"fi_rdm_tagged_pingpong -I 5 -v -U"
 	"fi_rdm_tagged_bw -I 5"
+	"fi_rdm_tagged_bw -I 5 -U"
 	"fi_rdm_tagged_bw -I 5 -v"
+	"fi_rdm_tagged_bw -I 5 -v -U"
 	"fi_dgram_pingpong -I 5"
 )
 
@@ -175,27 +189,40 @@ standard_tests=(
 	"fi_rma_bw -e msg -o read"
 	"fi_rma_bw -e msg -o writedata"
 	"fi_rma_bw -e rdm -o write"
+	"fi_rma_bw -e rdm -o write -U"
 	"fi_rma_bw -e rdm -o read"
+	"fi_rma_bw -e rdm -o read -U"
 	"fi_rma_bw -e rdm -o writedata"
+	"fi_rma_bw -e rdm -o writedata -U"
 	"fi_rdm_atomic -o all -I 1000"
+	"fi_rdm_atomic -o all -I 1000 -U"
 	"fi_rdm_cntr_pingpong"
 	"fi_multi_recv -e rdm"
 	"fi_multi_recv -e msg"
 	"fi_rdm_pingpong"
+	"fi_rdm_pingpong -U"
 	"fi_rdm_pingpong -v"
+	"fi_rdm_pingpong -v -U"
 	"fi_rdm_pingpong -k"
+	"fi_rdm_pingpong -k -U"
 	"fi_rdm_pingpong -k -v"
+	"fi_rdm_pingpong -k -v -U"
 	"fi_rdm_tagged_pingpong"
+	"fi_rdm_tagged_pingpong -U"
 	"fi_rdm_tagged_pingpong -v"
+	"fi_rdm_tagged_pingpong -v -U"
 	"fi_rdm_tagged_bw"
+	"fi_rdm_tagged_bw -U"
 	"fi_rdm_tagged_bw -v"
+	"fi_rdm_tagged_bw -v -U"
 	"fi_dgram_pingpong"
 	"fi_dgram_pingpong -k"
 )
 
 unit_tests=(
 	"fi_getinfo_test -s SERVER_ADDR GOOD_ADDR"
-	"fi_av_test -g GOOD_ADDR -n 1 -s SERVER_ADDR"
+	"fi_av_test -g GOOD_ADDR -n 1 -s SERVER_ADDR -e rdm"
+	"fi_av_test -g GOOD_ADDR -n 1 -s SERVER_ADDR -e dgram"
 	"fi_dom_test -n 2"
 	"fi_eq_test"
 	"fi_cq_test"
@@ -203,6 +230,10 @@ unit_tests=(
 	"fi_cntr_test"
 )
 
+regression_tests=(
+	"sighandler_test"
+)
+
 complex_tests=(
 	"fi_ubertest"
 )
@@ -213,6 +244,36 @@ multinode_tests=(
 	"fi_multinode_coll"
 )
 
+prov_efa_tests=( \
+	"fi_efa_rnr_read_cq_error"
+	"fi_efa_rnr_queue_resend -c 0 -S 1048576"
+	"fi_efa_rnr_queue_resend -c 0 -o read -S 4"
+	"fi_efa_rnr_queue_resend -c 0 -A read -S 4"
+	"fi_efa_rnr_queue_resend -c 0 -U -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -T -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -S 16384"
+	"fi_efa_rnr_queue_resend -c 1 -T -S 16384"
+	"fi_efa_rnr_queue_resend -c 1 -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -T -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -o write -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -o write -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -o read -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -o read -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -A write -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -A read -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -A cswap -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -U -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -T -U -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -U -S 16384"
+	"fi_efa_rnr_queue_resend -c 1 -T -U -S 16384"
+	"fi_efa_rnr_queue_resend -c 1 -U -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -T -U -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -o write -U -S 4"
+	"fi_efa_rnr_queue_resend -c 1 -o write -U -S 1048576"
+	"fi_efa_rnr_queue_resend -c 1 -A write -U -S 4"
+)
+
 function errcho {
 	>&2 echo $*
 }
@@ -253,6 +314,7 @@ function print_results {
 		esac
 
 		printf -- "- name:   %s\n" "$test_exe"
+		printf -- "  timestamp: %s\n" "$(date -u +'%Y%m%d-%H%M%S%z')"
 		printf -- "  result: %s\n" "$test_result"
 		printf -- "  time:   %s\n" "$test_time"
 		if [ $emit_stdout -eq 1 -a "$server_out_file" != "" ] ; then
@@ -427,6 +489,11 @@ function cs_test {
 	local start_time
 	local end_time
 	local test_time
+	local pin_core
+
+	if [ $# -eq 2 ]; then
+		pin_core=$2
+	fi
 
 	is_excluded "$test" && return
 
@@ -437,7 +504,7 @@ function cs_test {
 	else
 		s_arg="-s $S_INTERFACE"
 	fi
-	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} $s_arg"
+	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} ${pin_core} $s_arg"
 	${SERVER_CMD} "${EXPORT_ENV} $s_cmd" &> $s_outp &
 	s_pid=$!
 	sleep 1
@@ -447,14 +514,19 @@ function cs_test {
 	else
 		c_arg="-s $C_INTERFACE $S_INTERFACE"
 	fi
-	c_cmd="${BIN_PATH}${test_exe} ${C_ARGS} $c_arg"
+	c_cmd="${BIN_PATH}${test_exe} ${C_ARGS} ${pin_core} $c_arg"
 	${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_outp &
 	c_pid=$!
 
 	wait $c_pid
 	c_ret=$?
 
-	[[ c_ret -ne 0 ]] && kill -9 $s_pid 2> /dev/null
+	if [[ $c_ret -ne 0 ]] && ps -p $s_pid > /dev/null; then
+	    if [[ $STRICT_MODE -eq 0 ]]; then
+	        sleep 2
+	    fi
+	    kill -9 $s_pid 2> /dev/null
+	fi
 
 	wait $s_pid
 	s_ret=$?
@@ -483,6 +555,10 @@ function set_cfg_file {
 	local parent=$UTIL
 	local name=$CORE
 
+	if [[ ! -z "$COMPLEX_CFG" ]]; then
+		return
+	fi
+
 	if [ -z $UTIL ]; then
 		parent=$CORE
 		name=$1
@@ -510,11 +586,13 @@ function complex_test {
 	local end_time
 	local test_time
 
-	is_excluded "$test" && return
+	set_cfg_file $config
 	if [[ -z "$COMPLEX_CFG" ]]; then
-		set_cfg_file $config
+		return
 	fi
 
+	is_excluded "$test" && return
+
 	start_time=$(date '+%s')
 
 	if [[ $FORK -eq 1 ]]; then
@@ -527,12 +605,12 @@ function complex_test {
 		opts+=" -E"
 	fi
 
-	s_cmd="${BIN_PATH}${test_exe} -x $opts"
+	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} -x $opts"
 	FI_LOG_LEVEL=error ${SERVER_CMD} "${EXPORT_ENV} $s_cmd" &> $s_outp &
 	s_pid=$!
 	sleep 1
 
-	c_cmd="${BIN_PATH}${test_exe} -u "${COMPLEX_CFG}" $S_INTERFACE $opts"
+	c_cmd="${BIN_PATH}${test_exe} ${C_ARGS} -u "${COMPLEX_CFG}" $S_INTERFACE $opts"
 	FI_LOG_LEVEL=error ${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_outp &
 	c_pid=$!
 
@@ -580,7 +658,7 @@ function multinode_test {
 	local c_ret=0
 	local c_out_arr=()
 	local num_procs=$2
-	local test_exe="${test} -n $num_procs -p \"${PROV}\"" 	
+	local test_exe="${test} -n $num_procs -p \"${PROV}\""
 	local c_out
 	local start_time
 	local end_time
@@ -589,18 +667,18 @@ function multinode_test {
 	is_excluded "$test" && return
 
 	start_time=$(date '+%s')
-	
+
 	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} -s ${S_INTERFACE}"
 	${SERVER_CMD} "${EXPORT_ENV} $s_cmd" &> $s_outp &
 	s_pid=$!
 	sleep 1
-	
-	c_pid_arr=()	
+
+	c_pid_arr=()
 	for ((i=1; i<num_procs; i++))
 	do
 		local c_out=$(mktemp fabtests.c_outp${i}.XXXXXX)
 		c_cmd="${BIN_PATH}${test_exe} ${S_ARGS} -s ${S_INTERFACE}"
-		${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_out & 
+		${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_out &
 		c_pid_arr+=($!)
 		c_out_arr+=($c_out)
 	done
@@ -609,21 +687,26 @@ function multinode_test {
 		wait $pid
 		c_ret=($?)||$c_ret
 	done
-	
-	[[ c_ret -ne 0 ]] && kill -9 $s_pid 2> /dev/null
+
+	if [[ $c_ret -ne 0 ]] && ps -p $s_pid > /dev/null; then
+	    if [[ $STRICT_MODE -eq 0 ]]; then
+	        sleep 2
+	    fi
+	    kill -9 $s_pid 2> /dev/null
+	fi
 
 	wait $s_pid
 	s_ret=$?
 	echo "server finished"
-	
+
 	end_time=$(date '+%s')
 	test_time=$(compute_duration "$start_time" "$end_time")
-	
+
 	pe=1
 	if [[ $STRICT_MODE -eq 0 && $s_ret -eq $FI_ENODATA && $c_ret -eq $FI_ENODATA ]] ||
 	   [[ $STRICT_MODE -eq 0 && $s_ret -eq $FI_ENOSYS && $c_ret -eq $FI_ENOSYS ]]; then
 		print_results "$test_exe" "Notrun" "$test_time" "$s_outp" "$s_cmd" "" "$c_cmd"
-		for c_out in "${c_out_arr[@]}" 
+		for c_out in "${c_out_arr[@]}"
 		do
 			printf -- "  client_stdout $pe: |\n"
 			sed -e 's/^/    /' < $c_out
@@ -632,7 +715,8 @@ function multinode_test {
 		skip_count+=1
 	elif [ $s_ret -ne 0 -o $c_ret -ne 0 ]; then
 		print_results "$test_exe" "Fail" "$test_time" "$s_outp" "$s_cmd" "" "$c_cmd"
-		for c_out in "${c_out_arr[@]}" 
+		printf -- "  client_cmd: %s\n" "$c_cmd"
+		for c_out in "${c_out_arr[@]}"
 		do
 			printf -- "  client_stdout $pe: |\n"
 			sed -e 's/^/    /' < $c_out
@@ -644,7 +728,8 @@ function multinode_test {
 		fail_count+=1
 	else
 		print_results "$test_exe" "Pass" "$test_time" "$s_outp" "$s_cmd" "" "$c_cmd"
-		for c_out in "${c_out_arr[@]}" 
+		printf -- "  client_cmd: %s\n" "$c_cmd"
+		for c_out in "${c_out_arr[@]}"
 		do
 			printf -- "  client_stdout $pe: |\n"
 			sed -e 's/^/    /' < $c_out
@@ -654,6 +739,12 @@ function multinode_test {
 	fi
 }
 
+function prov_efa_test {
+	for test in "${prov_efa_tests[@]}"; do
+		cs_test "$test"
+	done
+}
+
 function set_core_util {
 	prov_arr=$(echo $PROV | tr ";" " ")
 	CORE=""
@@ -675,7 +766,7 @@ function main {
 
 	set_core_util
 	set_excludes
-	
+
 
 	if [[ $1 == "quick" ]]; then
 		local -r tests="unit functional short"
@@ -683,7 +774,7 @@ function main {
 		local -r tests="complex"
 		complex_type=$1
 	else
-		local -r tests=$(echo $1 | sed 's/all/unit,functional,standard,complex,multinode/g' | tr ',' ' ')
+		local -r tests=$(echo $1 | sed 's/all/unit,regression,functional,standard,complex,multinode/g' | tr ',' ' ')
 		if [[ $1 == "all" || $1 == "complex" ]]; then
 			complex_type="all"
 		fi
@@ -707,6 +798,11 @@ function main {
 				done
 			fi
 		;;
+		regression)
+			for test in "${regression_tests[@]}"; do
+				unit_test "$test" "0"
+			done
+		;;
 		functional)
 			for test in "${functional_tests[@]}"; do
 				cs_test "$test"
@@ -719,7 +815,11 @@ function main {
 		;;
 		standard)
 			for test in "${standard_tests[@]}"; do
-				cs_test "$test"
+				if [ ! -z $PIN_CORE ]; then
+					cs_test "$test" "--pin-core $PIN_CORE"
+				else
+					cs_test "$test"
+				fi
 			done
 		;;
 		complex)
@@ -739,6 +839,10 @@ function main {
 	esac
 	done
 
+	if [[ $PROVIDER_TESTS -eq 1 ]]; then
+		prov_${PROV}_test
+	fi
+
 	total=$(( $pass_count + $fail_count ))
 
 	print_border
@@ -774,6 +878,7 @@ function usage {
 			 regex patterns e.g. \"dgram,rma.*write\""
 	errcho -e " -E\texport provided variable name and value to ssh client and server processes.
 			 options must of of the form '-E var=value'"
+	errcho -e " -U\trun fabtests with FI_DELIVERY_COMPLETE set"
 	errcho -e " -f\texclude tests file: File containing list of test names /
 			 regex patterns to exclude (one per line)"
 	errcho -e " -N\tskip negative unit tests"
@@ -786,65 +891,83 @@ function usage {
 	errcho -e " -C\tAdditional client test arguments: Parameters to pass to client fabtests"
 	errcho -e " -L\tAdditional server test arguments: Parameters to pass to server fabtests"
 	errcho -e " -b\tenable out-of-band address exchange over the default port"
+	errcho -e " -P\tRun provider specific tests"
+	errcho -e " --pin-core\tSpecify cores to pin when running standard tests. Cores can specified via a comma-delimited list, e.g. 0,2-4"
 	exit 1
 }
 
-while getopts ":vt:p:g:e:f:c:s:u:T:C:L:NRSbkE:" opt; do
-case ${opt} in
-	t) TEST_TYPE=$OPTARG
-	;;
-	v) VERBOSE+=1
-	;;
-	p) BIN_PATH="${OPTARG}/"
-	;;
-	g) GOOD_ADDR=${OPTARG}
-	;;
-	f) read_exclude_file ${OPTARG}
-	;;
-	e) [[ -z "$input_excludes" ]] && input_excludes=${OPTARG} || \
-		input_excludes="${input_excludes},${OPTARG}"
-	;;
-	c) C_INTERFACE=${OPTARG}
-	;;
-	s) S_INTERFACE=${OPTARG}
-	;;
-	u) COMPLEX_CFG=${OPTARG}
-	;;
-	T) TIMEOUT_VAL=${OPTARG}
-	;;
-	N) SKIP_NEG+=1
-	;;
-	R)
-	;;
-	S) STRICT_MODE=1
-	;;
-	b) OOB=1
-	;;
-	k) FORK=1
-	;;
-	C) C_ARGS="${OPTARG}"
-	;;
-	L) S_ARGS="${OPTARG}"
-	;;
-	E)
-	delimiter="="
-	value=${OPTARG#*$delimiter}
-	var=${OPTARG:0:$(( ${#OPTARG} - ${#value} - ${#delimiter} ))}
-	EXPORT_STRING="export $var=\"$value\""
-	if [[ -z $EXPORT_ENV ]] ; then
-		EXPORT_ENV="$EXPORT_STRING ;"
-	else
-		EXPORT_ENV="$EXPORT_ENV $EXPORT_STRING ;"
-	fi
-	;;
-	:|\?) usage
-	;;
-esac
+Options=$(getopt --options v,t:,p:,g:,e:,f:,c:,s:,u:,T:,C:,L:,N,R,S,b,k,P,E:h \
+		 --longoptions pin-core:,help \
+		 --quiet \
+		 -- "$@")
+
+eval set -- "$Options"
+
+while true; do
+	case "$1" in
+		-t)
+		    TEST_TYPE=$2; shift 2 ;;
+		-v)
+		    VERBOSE+=1; shift ;;
+		-p)
+		    BIN_PATH="$2/"; shift 2 ;;
+		-g)
+		    GOOD_ADDR=$2; shift 2 ;;
+		-f)
+		    read_exclude_file $2; shift 2 ;;
+		-e)
+		    [[ -z "$input_excludes" ]] && input_excludes=${2} || \
+		    input_excludes="${input_excludes},${2}"
+		    shift 2 ;;
+		-c)
+		    C_INTERFACE=$2; shift 2 ;;
+		-s)
+		    S_INTERFACE=$2; shift 2 ;;
+		-u)
+		    COMPLEX_CFG=$2; shift 2 ;;
+		-T)
+		    TIMEOUT_VAL=$2; shift 2 ;;
+		-N)
+		    SKIP_NEG+=1; shift ;;
+		-P)
+		    PROVIDER_TESTS=1; shift ;;
+		-R)
+		    shift ;;
+		-S)
+		    STRICT_MODE=1; shift ;;
+		-b)
+		    OOB=1; shift ;;
+		-k)
+		    FORK=1; shift ;;
+		-C)
+		    C_ARGS=$2; shift 2 ;;
+		-L)
+		    S_ARGS=$2; shift 2 ;;
+		-E)
+		    delimiter="="
+		    value=${2#*$delimiter}
+		    var=${2:0:$(( ${#2} - ${#value} - ${#delimiter} ))}
+		    EXPORT_STRING="export $var=\"$value\""
+		    if [[ -z $EXPORT_ENV ]] ; then
+			EXPORT_ENV="$EXPORT_STRING ;"
+		    else
+			EXPORT_ENV="$EXPORT_ENV $EXPORT_STRING ;"
+		    fi
+		    shift 2 ;;
+		--pin-core)
+		    PIN_CORE=$2; shift 2 ;;
+		-h | --help)
+		    usage ;;
+		--)
+		    shift ; break ;;
+		*)
+		    usage ;;
+	esac
 
 done
 
 # base ssh command
-declare bssh="ssh -n -o StrictHostKeyChecking=no -o ConnectTimeout=2 -o BatchMode=yes"
+declare bssh="ssh -n -o StrictHostKeyChecking=no -o ConnectTimeout=30 -o BatchMode=yes"
 if [ -z "$(which timeout 2> /dev/null)" ]; then
 	# forego timeout
 	declare SERVER_CMD="eval"
