@@ -2,6 +2,7 @@
  * Copyright (c) 2013-2018 Intel Corporation. All rights reserved.
  * Copyright (c) 2016-2018 Cisco Systems, Inc. All rights reserved.
  * Copyright (c) 2019 Amazon.com, Inc. or its affiliates. All rights reserved.
+ * Copyright (c) 2022 DataDirect Networks, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -35,7 +36,7 @@
 #ifndef _OFI_H_
 #define _OFI_H_
 
-#include "config.h"
+#include <config.h>
 
 #include <assert.h>
 #include <pthread.h>
@@ -50,7 +51,6 @@
 #include <ofi_file.h>
 #include <ofi_lock.h>
 #include <ofi_atom.h>
-#include <ofi_mem.h>
 #include <ofi_net.h>
 #include <rdma/providers/fi_prov.h>
 #include <rdma/providers/fi_log.h>
@@ -75,6 +75,7 @@ extern "C" {
 #define OFI_GETINFO_INTERNAL	(1ULL << 58)
 #define OFI_CORE_PROV_ONLY	(1ULL << 59)
 #define OFI_GETINFO_HIDDEN	(1ULL << 60)
+#define OFI_OFFLOAD_PROV_ONLY	(1ULL << 61)
 
 #define OFI_ORDER_RAR_SET	(FI_ORDER_RAR | FI_ORDER_RMA_RAR | \
 				 FI_ORDER_ATOMIC_RAR)
@@ -85,6 +86,37 @@ extern "C" {
 #define OFI_ORDER_WAW_SET	(FI_ORDER_WAW | FI_ORDER_RMA_WAW | \
 				 FI_ORDER_ATOMIC_WAW)
 
+#define OFI_PRIMARY_TX_CAPS \
+	(FI_MSG | FI_RMA | FI_TAGGED | FI_ATOMIC | FI_MULTICAST | \
+	 FI_READ | FI_WRITE | FI_SEND | \
+	 FI_COLLECTIVE | FI_NAMED_RX_CTX | FI_HMEM)
+
+#define OFI_SECONDARY_TX_CAPS \
+	(FI_TRIGGER | FI_FENCE | FI_RMA_PMEM)
+
+#define OFI_PRIMARY_RX_CAPS \
+	(FI_MSG | FI_RMA | FI_TAGGED | FI_ATOMIC | \
+	 FI_REMOTE_READ | FI_REMOTE_WRITE | FI_RECV | \
+	 FI_DIRECTED_RECV | FI_VARIABLE_MSG | \
+	 FI_COLLECTIVE | FI_HMEM)
+
+#define OFI_SECONDARY_RX_CAPS \
+	(FI_MULTI_RECV | FI_TRIGGER | FI_RMA_PMEM | FI_SOURCE | \
+	 FI_RMA_EVENT | FI_SOURCE_ERR)
+
+#define OFI_PRIMARY_CAPS \
+	(OFI_PRIMARY_TX_CAPS | OFI_PRIMARY_RX_CAPS | \
+	 FI_REMOTE_COMM | FI_LOCAL_COMM)
+
+#define OFI_SECONDARY_CAPS \
+	(OFI_SECONDARY_TX_CAPS | OFI_SECONDARY_RX_CAPS | \
+	 FI_SHARED_AV)
+
+#define OFI_TX_MSG_CAPS (FI_MSG | FI_SEND)
+#define OFI_RX_MSG_CAPS (FI_MSG | FI_RECV)
+#define OFI_TX_RMA_CAPS (FI_RMA | FI_READ | FI_WRITE)
+#define OFI_RX_RMA_CAPS (FI_RMA | FI_REMOTE_READ | FI_REMOTE_WRITE)
+
 #define OFI_IGNORED_TX_CAPS /* older Rx caps not applicable to Tx */ \
 	(FI_REMOTE_READ | FI_REMOTE_WRITE | FI_RECV | FI_DIRECTED_RECV | \
 	 FI_VARIABLE_MSG | FI_MULTI_RECV | FI_SOURCE | FI_RMA_EVENT | \
@@ -93,6 +125,13 @@ extern "C" {
 	(FI_READ | FI_WRITE | FI_SEND | FI_FENCE | FI_MULTICAST | \
 	 FI_NAMED_RX_CTX)
 
+#define OFI_TX_OP_FLAGS \
+	(FI_COMMIT_COMPLETE | FI_COMPLETION | FI_DELIVERY_COMPLETE | \
+	 FI_INJECT | FI_INJECT_COMPLETE | FI_MULTICAST | FI_TRANSMIT_COMPLETE)
+
+#define OFI_RX_OP_FLAGS \
+	(FI_COMPLETION | FI_MULTI_RECV)
+
 
 #define sizeof_field(type, field) sizeof(((type *)0)->field)
 
@@ -133,8 +172,6 @@ static inline int ofi_val32_ge(uint32_t x, uint32_t y) {
 #define ofi_val64_inrange(start, length, value) \
     ofi_val64_ge(value, start) && ofi_val64_lt(value, start + length)
 
-#define OFI_MAGIC_64 (0x0F1C0DE0F1C0DE64)
-
 #ifndef BIT
 #define BIT(nr) (1UL << (nr))
 #endif
@@ -155,30 +192,31 @@ static inline int ofi_val32_ge(uint32_t x, uint32_t y) {
 #define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))
 #endif
 
+#ifndef STRINGIFY
+#define __STRINGIFY(expr) #expr
+#define STRINGIFY(expr) __STRINGIFY(expr)
+#endif
+
 #define TAB "    "
 
-#define CASEENUMSTR(SYM) \
-	case SYM: { ofi_strcatf(buf, #SYM); break; }
-#define IFFLAGSTR(flags, SYM) \
-	do { if (flags & SYM) ofi_strcatf(buf, #SYM ", "); } while(0)
 #define CASEENUMSTRN(SYM, N) \
 	case SYM: { ofi_strncatf(buf, N, #SYM); break; }
 #define IFFLAGSTRN(flags, SYM, N) \
 	do { if (flags & SYM) ofi_strncatf(buf, N, #SYM ", "); } while(0)
 
-#define ofi_strcatf(dest, ...) \
-	ofi_strncatf(dest, OFI_BUFSIZ, __VA_ARGS__)
 
 /*
  * CPU specific features
  */
+
+/* X86_64 */
 enum {
-	OFI_CLWB_REG		= 2,
+	OFI_CLWB_REG		= 1,
 	OFI_CLWB_BIT		= (1 << 24),
 	OFI_CLFLUSHOPT_REG	= 1,
-	OFI_CLFLUSHOPT_BIT	= (1 << 24),
+	OFI_CLFLUSHOPT_BIT	= (1 << 23),
 	OFI_CLFLUSH_REG		= 3,
-	OFI_CLFLUSH_BIT		= (1 << 23),
+	OFI_CLFLUSH_BIT		= (1 << 19),
 };
 
 int ofi_cpu_supports(unsigned func, unsigned reg, unsigned bit);
@@ -188,41 +226,44 @@ enum ofi_prov_type {
 	OFI_PROV_CORE,
 	OFI_PROV_UTIL,
 	OFI_PROV_HOOK,
+	OFI_PROV_OFFLOAD,
 };
 
 /* Restrict to size of struct fi_provider::context (struct fi_context) */
-struct fi_prov_context {
+struct ofi_prov_context {
 	enum ofi_prov_type type;
-	int disable_logging;
+	bool disable_logging;
+	bool disable_layering;	/* applies to core providers only */
 };
 
-struct fi_filter {
+static inline struct ofi_prov_context *
+ofi_prov_ctx(const struct fi_provider *prov)
+{
+	return (struct ofi_prov_context *) &prov->context;
+}
+
+struct ofi_filter {
 	char **names;
-	int negated;
+	bool negated;
 };
 
-extern struct fi_filter prov_log_filter;
+extern struct ofi_filter prov_log_filter;
 extern struct fi_provider core_prov;
+extern const char *log_prefix;
 
-void ofi_create_filter(struct fi_filter *filter, const char *env_name);
-void ofi_free_filter(struct fi_filter *filter);
-int ofi_apply_filter(struct fi_filter *filter, const char *name);
+void ofi_create_filter(struct ofi_filter *filter, const char *env_name);
+void ofi_free_filter(struct ofi_filter *filter);
+int ofi_apply_filter(struct ofi_filter *filter, const char *name);
 
 int ofi_nic_close(struct fid *fid);
 struct fid_nic *ofi_nic_dup(const struct fid_nic *nic);
 int ofi_nic_tostr(const struct fid *fid_nic, char *buf, size_t len);
 
-struct fi_provider *ofi_get_hook(const char *name);
-
 void fi_log_init(void);
 void fi_log_fini(void);
 void fi_param_init(void);
 void fi_param_fini(void);
 void fi_param_undefine(const struct fi_provider *provider);
-void ofi_hook_init(void);
-void ofi_hook_fini(void);
-void ofi_hook_install(struct fid_fabric *hfabric, struct fid_fabric **fabric,
-		      struct fi_provider *prov);
 void ofi_remove_comma(char *buffer);
 void ofi_strncatf(char *dest, size_t n, const char *fmt, ...);
 
@@ -262,14 +303,20 @@ static inline void *ofi_get_page_start(const void *addr, size_t page_size)
 
 static inline void *ofi_get_page_end(const void *addr, size_t page_size)
 {
-	return ofi_get_page_start((const char *) addr + page_size -1, page_size);
+	return (void *)((uintptr_t)ofi_get_page_start((const char *)addr
+			+ page_size, page_size) - 1);
 }
 
 static inline size_t
 ofi_get_page_bytes(const void *addr, size_t len, size_t page_size)
 {
-	return (char *)ofi_get_page_end((const char *) addr + len, page_size) -
-	       (char *)ofi_get_page_start(addr, page_size);
+	char *start = (char *)ofi_get_page_start(addr, page_size);
+	char *end = (char *)ofi_get_page_start((const char*)addr + len - 1, page_size)
+		    + page_size;
+	size_t result = end - start;
+
+	assert(result % page_size == 0);
+	return result;
 }
 
 #define FI_TAG_GENERIC	0xAAAAAAAAAAAAAAAAULL
@@ -281,11 +328,15 @@ uint8_t ofi_msb(uint64_t num);
 uint8_t ofi_lsb(uint64_t num);
 
 extern size_t ofi_universe_size;
+extern int ofi_av_remove_cleanup;
+
+bool ofi_send_allowed(uint64_t caps);
+bool ofi_recv_allowed(uint64_t caps);
+bool ofi_rma_initiate_allowed(uint64_t caps);
+bool ofi_rma_target_allowed(uint64_t caps);
+bool ofi_needs_tx(uint64_t caps);
+bool ofi_needs_rx(uint64_t caps);
 
-int ofi_send_allowed(uint64_t caps);
-int ofi_recv_allowed(uint64_t caps);
-int ofi_rma_initiate_allowed(uint64_t caps);
-int ofi_rma_target_allowed(uint64_t caps);
 int ofi_ep_bind_valid(const struct fi_provider *prov, struct fid *bfid,
 		      uint64_t flags);
 int ofi_check_rx_mode(const struct fi_info *info, uint64_t flags);
@@ -375,6 +426,11 @@ uint32_t ofi_generate_seed(void);
 
 size_t ofi_vrb_speed(uint8_t speed, uint8_t width);
 
+int ofi_open_log(uint32_t version, void *attr, size_t attr_len,
+		 uint64_t flags, struct fid **fid, void *context);
+void ofi_tostr_log_level(char *buf, size_t len, enum fi_log_level level);
+void ofi_tostr_log_subsys(char *buf, size_t len, enum fi_log_subsys subsys);
+
 #ifdef __cplusplus
 }
 #endif
