@@ -56,12 +56,17 @@ static struct fi_ops X = {
 	.bind = fi_no_bind,
 	.control = fi_no_control,
 	.ops_open = fi_no_ops_open,
+	.tostr = fi_no_ops_tostr,
+	.ops_set = fi_no_ops_set,
 };
  */
 int fi_no_bind(struct fid *fid, struct fid *bfid, uint64_t flags);
 int fi_no_control(struct fid *fid, int command, void *arg);
 int fi_no_ops_open(struct fid *fid, const char *name,
 		uint64_t flags, void **ops, void *context);
+int fi_no_tostr(const struct fid *fid, char *buf, size_t len);
+int fi_no_ops_set(struct fid *fid, const char *name, uint64_t flags,
+		  void *ops, void *context);
 
 /*
 static struct fi_ops_fabric X = {
@@ -71,10 +76,13 @@ static struct fi_ops_fabric X = {
 	.eq_open = fi_no_eq_open,
 	.wait_open = fi_no_wait_open,
 	.trywait = fi_no_trywait,
+	.domain2 = fi_no_domain2,
 };
 */
 int fi_no_domain(struct fid_fabric *fabric, struct fi_info *info,
 		struct fid_domain **dom, void *context);
+int fi_no_domain2(struct fid_fabric *fabric, struct fi_info *info,
+		struct fid_domain **dom, uint64_t flags, void *context);
 int fi_no_passive_ep(struct fid_fabric *fabric, struct fi_info *info,
 		struct fid_pep **pep, void *context);
 int fi_no_eq_open(struct fid_fabric *fabric, struct fi_eq_attr *attr,
@@ -192,6 +200,7 @@ static struct fi_ops_domain X = {
 	.srx_ctx = fi_no_srx_context,
 	.query_atomic = fi_no_query_atomic,
 	.query_collective = fi_no_query_collective,
+	.endpoint2 = fi_no_endpoint2,
 };
 */
 int fi_no_av_open(struct fid_domain *domain, struct fi_av_attr *attr,
@@ -200,6 +209,8 @@ int fi_no_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 		struct fid_cq **cq, void *context);
 int fi_no_endpoint(struct fid_domain *domain, struct fi_info *info,
 		struct fid_ep **ep, void *context);
+int fi_no_endpoint2(struct fid_domain *domain, struct fi_info *info,
+		struct fid_ep **ep, uint64_t flags, void *context);
 int fi_no_scalable_ep(struct fid_domain *domain, struct fi_info *info,
 		struct fid_ep **sep, void *context);
 int fi_no_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
@@ -457,6 +468,25 @@ int fi_no_av_insertsym(struct fid_av *av, const char *node, size_t nodecnt,
 int fi_no_av_remove(struct fid_av *av, fi_addr_t *fi_addr, size_t count,
 			uint64_t flags);
 
+/*
+static struct fi_ops_av_set X = {
+	.size = sizeof(struct fi_ops_av),
+	.set_union = fi_no_av_set_union,
+	.intersect = fi_no_av_set_intersect,
+	.diff = fi_no_av_set_diff,
+	.insert = fi_no_av_set_insert,
+	.remove = fi_no_av_set_remove,
+	.straddr = X,
+};
+*/
+
+int fi_no_av_set_union(struct fid_av_set *dst, const struct fid_av_set *src);
+int fi_no_av_set_intersect(struct fid_av_set *dst, const struct fid_av_set *src);
+int fi_no_av_set_diff(struct fid_av_set *dst, const struct fid_av_set *src);
+int fi_no_av_set_insert(struct fid_av_set *set, fi_addr_t addr);
+int fi_no_av_set_remove(struct fid_av_set *set, fi_addr_t addr);
+int fi_no_av_set_addr(struct fid_av_set *set, fi_addr_t *coll_addr);
+
 /*
 static struct fi_ops_collective X = {
 	.size = sizeof(struct fi_ops_collective),
@@ -473,6 +503,8 @@ static struct fi_ops_collective X = {
 };
 */
 ssize_t fi_coll_no_barrier(struct fid_ep *ep, fi_addr_t coll_addr, void *context);
+ssize_t fi_coll_no_barrier2(struct fid_ep *ep, fi_addr_t coll_addr, uint64_t flags,
+			    void *context);
 ssize_t fi_coll_no_broadcast(struct fid_ep *ep, void *buf, size_t count, void *desc,
 			     fi_addr_t coll_addr, fi_addr_t root_addr,
 			     enum fi_datatype datatype, uint64_t flags, void *context);
