@@ -1,6 +1,7 @@
 /*
  * Copyright (c) 2017-2019 Intel Corporation, Inc. All rights reserved.
- * Copyright (c) 2019 Amazon.com, Inc. or its affiliates. All rights reserved.
+ * Copyright (c) 2019-2021 Amazon.com, Inc. or its affiliates.
+ *                         All rights reserved.
  * (C) Copyright 2020 Hewlett Packard Enterprise Development LP
  *
  * This software is available to you under a choice of one of two
@@ -49,10 +50,16 @@
 #include <ofi_tree.h>
 #include <ofi_hmem.h>
 
+
+int ofi_open_mr_cache(uint32_t version, void *attr, size_t attr_len,
+		      uint64_t flags, struct fid **fid, void *context);
+
 struct ofi_mr_info {
 	struct iovec iov;
 	enum fi_hmem_iface iface;
 	uint64_t device;
+	void     *ipc_mapped_addr;
+	uint8_t  ipc_handle[MAX_IPC_HANDLE_SIZE];
 };
 
 
@@ -102,6 +109,14 @@ static inline uint64_t ofi_mr_get_prov_mode(uint32_t version,
 
 /* Single lock used by all memory monitors and MR caches. */
 extern pthread_mutex_t mm_lock;
+/* The read-write lock is an additional lock used to protect the dlist_entry
+ * list of ofi_mem_monitor. Due to the necessity of releasing the mm_lock
+ * while walking the dlist in ofi_monitor_notify, we need a separate lock to
+ * ensure thread safety. This must be a read-write lock because
+ * ofi_monitor_notify may be recursive and cannot block multiple walks from
+ * occurring at the same time.
+ */
+extern pthread_rwlock_t mm_list_rwlock;
 
 /*
  * Memory notifier - Report memory mapping changes to address ranges
@@ -111,11 +126,30 @@ struct ofi_mr_cache;
 
 union ofi_mr_hmem_info {
 	uint64_t cuda_id;
+	uint64_t ze_id;
+};
+
+struct ofi_mr_entry {
+	struct ofi_mr_info		info;
+	struct ofi_rbnode		*node;
+	int				use_cnt;
+	struct dlist_entry		list_entry;
+	union ofi_mr_hmem_info		hmem_info;
+	uint8_t				data[];
+};
+
+enum fi_mm_state {
+	FI_MM_STATE_UNSPEC = 0,
+	FI_MM_STATE_IDLE,
+	FI_MM_STATE_STARTING,
+	FI_MM_STATE_RUNNING,
+	FI_MM_STATE_STOPPING,
 };
 
 struct ofi_mem_monitor {
 	struct dlist_entry		list;
 	enum fi_hmem_iface		iface;
+	enum fi_mm_state                state;
 
 	void (*init)(struct ofi_mem_monitor *monitor);
 	void (*cleanup)(struct ofi_mem_monitor *monitor);
@@ -134,19 +168,22 @@ struct ofi_mem_monitor {
 	 * pages behind a given virtual address have changed), the buffer needs
 	 * to be re-registered.
 	 */
-	bool (*valid)(struct ofi_mem_monitor *notifier, const void *addr,
-		      size_t len, union ofi_mr_hmem_info *hmem_info);
+	bool (*valid)(struct ofi_mem_monitor *notifier,
+		      const struct ofi_mr_info *info, struct ofi_mr_entry *entry);
 };
 
 void ofi_monitor_init(struct ofi_mem_monitor *monitor);
 void ofi_monitor_cleanup(struct ofi_mem_monitor *monitor);
 void ofi_monitors_init(void);
 void ofi_monitors_cleanup(void);
+int ofi_monitor_import(struct fid *fid);
+
 int ofi_monitors_add_cache(struct ofi_mem_monitor **monitors,
 			   struct ofi_mr_cache *cache);
 void ofi_monitors_del_cache(struct ofi_mr_cache *cache);
 void ofi_monitor_notify(struct ofi_mem_monitor *monitor,
 			const void *addr, size_t len);
+void ofi_monitor_flush(struct ofi_mem_monitor *monitor);
 
 int ofi_monitor_subscribe(struct ofi_mem_monitor *monitor,
 			  const void *addr, size_t len,
@@ -155,9 +192,18 @@ void ofi_monitor_unsubscribe(struct ofi_mem_monitor *monitor,
 			     const void *addr, size_t len,
 			     union ofi_mr_hmem_info *hmem_info);
 
+int ofi_monitor_start_no_op(struct ofi_mem_monitor *monitor);
+void ofi_monitor_stop_no_op(struct ofi_mem_monitor *monitor);
+int ofi_monitor_subscribe_no_op(struct ofi_mem_monitor *notifier,
+				 const void *addr, size_t len,
+				 union ofi_mr_hmem_info *hmem_info);
+void ofi_monitor_unsubscribe_no_op(struct ofi_mem_monitor *notifier,
+				    const void *addr, size_t len,
+				    union ofi_mr_hmem_info *hmem_info);
 extern struct ofi_mem_monitor *default_monitor;
 extern struct ofi_mem_monitor *default_cuda_monitor;
 extern struct ofi_mem_monitor *default_rocr_monitor;
+extern struct ofi_mem_monitor *default_ze_monitor;
 
 /*
  * Userfault fd memory monitor
@@ -181,8 +227,10 @@ struct ofi_memhooks {
 extern struct ofi_mem_monitor *memhooks_monitor;
 
 extern struct ofi_mem_monitor *cuda_monitor;
-
+extern struct ofi_mem_monitor *cuda_ipc_monitor;
 extern struct ofi_mem_monitor *rocr_monitor;
+extern struct ofi_mem_monitor *ze_monitor;
+extern struct ofi_mem_monitor *import_monitor;
 
 /*
  * Used to store registered memory regions into a lookup map.  This
@@ -252,42 +300,12 @@ struct ofi_mr_cache_params {
 	char *				monitor;
 	int				cuda_monitor_enabled;
 	int				rocr_monitor_enabled;
+	int				ze_monitor_enabled;
 };
 
 extern struct ofi_mr_cache_params	cache_params;
 
-struct ofi_mr_entry {
-	struct ofi_mr_info		info;
-	void				*storage_context;
-	int				use_cnt;
-	struct dlist_entry		list_entry;
-	union ofi_mr_hmem_info		hmem_info;
-	uint8_t				data[];
-};
-
-enum ofi_mr_storage_type {
-	OFI_MR_STORAGE_DEFAULT = 0,
-	OFI_MR_STORAGE_RBT,
-	OFI_MR_STORAGE_USER,
-};
-
-struct ofi_mr_storage {
-	enum ofi_mr_storage_type	type;
-	void				*storage;
-
-	struct ofi_mr_entry *		(*find)(struct ofi_mr_storage *storage,
-						const struct ofi_mr_info *key);
-	struct ofi_mr_entry *		(*overlap)(struct ofi_mr_storage *storage,
-						const struct iovec *key);
-	int				(*insert)(struct ofi_mr_storage *storage,
-						struct ofi_mr_info *key,
-						struct ofi_mr_entry *entry);
-	int				(*erase)(struct ofi_mr_storage *storage,
-						struct ofi_mr_entry *entry);
-	void				(*destroy)(struct ofi_mr_storage *storage);
-};
-
-#define OFI_HMEM_MAX 4
+#define OFI_HMEM_MAX 6
 
 struct ofi_mr_cache {
 	struct util_domain		*domain;
@@ -295,9 +313,9 @@ struct ofi_mr_cache {
 	struct dlist_entry		notify_entries[OFI_HMEM_MAX];
 	size_t				entry_data_size;
 
-	struct ofi_mr_storage		storage;
+	struct ofi_rbmap		tree;
 	struct dlist_entry		lru_list;
-	struct dlist_entry		flush_list;
+	struct dlist_entry		dead_region_list;
 	pthread_mutex_t 		lock;
 
 	size_t				cached_cnt;
@@ -323,10 +341,36 @@ void ofi_mr_cache_cleanup(struct ofi_mr_cache *cache);
 
 void ofi_mr_cache_notify(struct ofi_mr_cache *cache, const void *addr, size_t len);
 
+int ofi_ipc_cache_open(struct ofi_mr_cache **cache,
+			struct util_domain *domain);
+void ofi_ipc_cache_destroy(struct ofi_mr_cache *cache);
+int  ofi_ipc_cache_search(struct ofi_mr_cache *cache, struct ipc_info *ipc_info,
+			   struct ofi_mr_entry **mr_entry);
+
+static inline bool ofi_mr_cache_full(struct ofi_mr_cache *cache)
+{
+	return (cache->cached_cnt >= cache_params.max_cnt) ||
+	       (cache->cached_size >= cache_params.max_size);
+}
+
 bool ofi_mr_cache_flush(struct ofi_mr_cache *cache, bool flush_lru);
 
-int ofi_mr_cache_search(struct ofi_mr_cache *cache, const struct fi_mr_attr *attr,
-			struct ofi_mr_entry **entry);
+/**
+ * @brief Given an ofi_mr_info (with an iov range, ipc_info)
+ * If the iov range is already registered and validated by the monitor,
+ * assign the corresponding ofi_mr_entry to entry. Otherwise, register
+ * a new ofi_mr_entry and assign it to entry.
+ *
+ * @param[in] cache     The cache the entry belongs to
+ * @param[in] info      Information about the mr entry to search
+ * @param[out] entry    The registered entry corresponding to the
+ *			region described in info.
+ * @returns On success, returns 0. On failure, returns a negative error code.
+ */
+int ofi_mr_cache_search(struct ofi_mr_cache *cache,
+			 const struct ofi_mr_info *info,
+			 struct ofi_mr_entry **entry);
+
 /**
  * Given an attr (with an iov range), if the iov range is already registered,
  * return the corresponding ofi_mr_entry. Otherwise, return NULL.
