@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
+ * (C) Copyright 2020 Hewlett Packard Enterprise Development LP
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -33,6 +34,7 @@
 #ifndef FI_DOMAIN_H
 #define FI_DOMAIN_H
 
+#include <string.h>
 #include <rdma/fabric.h>
 #include <rdma/fi_eq.h>
 
@@ -49,7 +51,16 @@ extern "C" {
 
 #define FI_SYMMETRIC		(1ULL << 59)
 #define FI_SYNC_ERR		(1ULL << 58)
-
+#define FI_UNIVERSE		(1ULL << 57)
+#define FI_BARRIER_SET		(1ULL << 40)
+#define FI_BROADCAST_SET	(1ULL << 41)
+#define FI_ALLTOALL_SET		(1ULL << 42)
+#define FI_ALLREDUCE_SET	(1ULL << 43)
+#define FI_ALLGATHER_SET	(1ULL << 44)
+#define FI_REDUCE_SCATTER_SET	(1ULL << 45)
+#define FI_REDUCE_SET		(1ULL << 46)
+#define FI_SCATTER_SET		(1ULL << 47)
+#define FI_GATHER_SET		(1ULL << 48)
 
 struct fi_av_attr {
 	enum fi_av_type		type;
@@ -61,6 +72,18 @@ struct fi_av_attr {
 	uint64_t		flags;
 };
 
+struct fi_av_set_attr {
+	size_t			count;
+	fi_addr_t		start_addr;
+	fi_addr_t		end_addr;
+	uint64_t		stride;
+	size_t			comm_key_size;
+	uint8_t			*comm_key;
+	uint64_t		flags;
+};
+
+struct fid_av_set;
+
 struct fi_ops_av {
 	size_t	size;
 	int	(*insert)(struct fid_av *av, const void *addr, size_t count,
@@ -77,6 +100,8 @@ struct fi_ops_av {
 			size_t *addrlen);
 	const char * (*straddr)(struct fid_av *av, const void *addr,
 			char *buf, size_t *len);
+	int	(*av_set)(struct fid_av *av, struct fi_av_set_attr *attr,
+			struct fid_av_set **av_set, void *context);
 };
 
 struct fid_av {
@@ -96,6 +121,15 @@ struct fid_mr {
 	uint64_t		key;
 };
 
+enum fi_hmem_iface {
+	FI_HMEM_SYSTEM	= 0,
+	FI_HMEM_CUDA,
+	FI_HMEM_ROCR,
+	FI_HMEM_ZE,
+	FI_HMEM_NEURON,
+	FI_HMEM_SYNAPSEAI,
+};
+
 struct fi_mr_attr {
 	const struct iovec	*mr_iov;
 	size_t			iov_count;
@@ -105,6 +139,14 @@ struct fi_mr_attr {
 	void			*context;
 	size_t			auth_key_size;
 	uint8_t			*auth_key;
+	enum fi_hmem_iface	iface;
+	union {
+		uint64_t	reserved;
+		int		cuda;
+		int		ze;
+		int		neuron;
+		int		synapseai;
+	} device;
 };
 
 struct fi_mr_modify {
@@ -112,6 +154,23 @@ struct fi_mr_modify {
 	struct fi_mr_attr	attr;
 };
 
+#define FI_SET_OPS_HMEM_OVERRIDE "hmem_override_ops"
+
+struct fi_hmem_override_ops {
+	size_t	size;
+
+	ssize_t	(*copy_from_hmem_iov)(void *dest, size_t size,
+				      enum fi_hmem_iface iface, uint64_t device,
+				      const struct iovec *hmem_iov,
+				      size_t hmem_iov_count,
+				      uint64_t hmem_iov_offset);
+
+	ssize_t (*copy_to_hmem_iov)(enum fi_hmem_iface iface, uint64_t device,
+				    const struct iovec *hmem_iov,
+				    size_t hmem_iov_count,
+				    uint64_t hmem_iov_offset, const void *src,
+				    size_t size);
+};
 
 #ifdef FABRIC_DIRECT
 #include <rdma/fi_direct_atomic_def.h>
@@ -119,6 +178,8 @@ struct fi_mr_modify {
 
 #ifndef FABRIC_DIRECT_ATOMIC_DEF
 
+#define FI_COLLECTIVE_OFFSET 256
+
 enum fi_datatype {
 	FI_INT8,
 	FI_UINT8,
@@ -134,7 +195,17 @@ enum fi_datatype {
 	FI_DOUBLE_COMPLEX,
 	FI_LONG_DOUBLE,
 	FI_LONG_DOUBLE_COMPLEX,
-	FI_DATATYPE_LAST
+	/* End of point to point atomic datatypes */
+	FI_DATATYPE_LAST,
+	/*
+	 * enums for 128-bit integer atomics, existing ordering and
+	 * FI_DATATYPE_LAST preserved for compatabilty.
+	 */
+	FI_INT128 = FI_DATATYPE_LAST,
+	FI_UINT128,
+
+	/* Collective datatypes */
+	FI_VOID = FI_COLLECTIVE_OFFSET,
 };
 
 enum fi_op {
@@ -157,7 +228,27 @@ enum fi_op {
 	FI_CSWAP_GE,
 	FI_CSWAP_GT,
 	FI_MSWAP,
-	FI_ATOMIC_OP_LAST
+	/* End of point to point atomic ops */
+	FI_ATOMIC_OP_LAST,
+
+	/* Collective datatypes */
+	FI_NOOP = FI_COLLECTIVE_OFFSET,
+};
+
+#endif
+
+#ifndef FABRIC_DIRECT_COLLECTIVE_DEF
+
+enum fi_collective_op {
+	FI_BARRIER,
+	FI_BROADCAST,
+	FI_ALLTOALL,
+	FI_ALLREDUCE,
+	FI_ALLGATHER,
+	FI_REDUCE_SCATTER,
+	FI_REDUCE,
+	FI_SCATTER,
+	FI_GATHER,
 };
 
 #endif
@@ -166,6 +257,7 @@ enum fi_op {
 struct fi_atomic_attr;
 struct fi_cq_attr;
 struct fi_cntr_attr;
+struct fi_collective_attr;
 
 struct fi_ops_domain {
 	size_t	size;
@@ -190,6 +282,11 @@ struct fi_ops_domain {
 	int	(*query_atomic)(struct fid_domain *domain,
 			enum fi_datatype datatype, enum fi_op op,
 			struct fi_atomic_attr *attr, uint64_t flags);
+	int	(*query_collective)(struct fid_domain *domain,
+			enum fi_collective_op coll,
+			struct fi_collective_attr *attr, uint64_t flags);
+	int	(*endpoint2)(struct fid_domain *domain, struct fi_info *info,
+			struct fid_ep **ep, uint64_t flags, void *context);
 };
 
 /* Memory registration flags */
@@ -231,6 +328,18 @@ fi_domain(struct fid_fabric *fabric, struct fi_info *info,
 	return fabric->ops->domain(fabric, info, domain, context);
 }
 
+static inline int
+fi_domain2(struct fid_fabric *fabric, struct fi_info *info,
+	   struct fid_domain **domain, uint64_t flags, void *context)
+{
+	if (!flags)
+		return fi_domain(fabric, info, domain, context);
+
+	return FI_CHECK_OP(fabric->ops, struct fi_ops_fabric, domain2) ?
+		fabric->ops->domain2(fabric, info, domain, flags, context) :
+		-FI_ENOSYS;
+}
+
 static inline int
 fi_domain_bind(struct fid_domain *domain, struct fid *fid, uint64_t flags)
 {
@@ -267,20 +376,20 @@ fi_poll_open(struct fid_domain *domain, struct fi_poll_attr *attr,
 
 static inline int
 fi_mr_reg(struct fid_domain *domain, const void *buf, size_t len,
-	  uint64_t access, uint64_t offset, uint64_t requested_key,
+	  uint64_t acs, uint64_t offset, uint64_t requested_key,
 	  uint64_t flags, struct fid_mr **mr, void *context)
 {
-	return domain->mr->reg(&domain->fid, buf, len, access, offset,
+	return domain->mr->reg(&domain->fid, buf, len, acs, offset,
 			       requested_key, flags, mr, context);
 }
 
 static inline int
 fi_mr_regv(struct fid_domain *domain, const struct iovec *iov,
-			size_t count, uint64_t access,
+			size_t count, uint64_t acs,
 			uint64_t offset, uint64_t requested_key,
 			uint64_t flags, struct fid_mr **mr, void *context)
 {
-	return domain->mr->regv(&domain->fid, iov, count, access,
+	return domain->mr->regv(&domain->fid, iov, count, acs,
 			offset, requested_key, flags, mr, context);
 }
 
@@ -305,12 +414,11 @@ static inline int
 fi_mr_raw_attr(struct fid_mr *mr, uint64_t *base_addr,
 	       uint8_t *raw_key, size_t *key_size, uint64_t flags)
 {
-	struct fi_mr_raw_attr attr = {
-		.flags = flags,
-		.base_addr = base_addr,
-		.raw_key = raw_key,
-		.key_size = key_size
-	};
+	struct fi_mr_raw_attr attr;
+	attr.flags = flags;
+	attr.base_addr = base_addr;
+	attr.raw_key = raw_key;
+	attr.key_size = key_size;
 	return mr->fid.ops->control(&mr->fid, FI_GET_RAW_MR, &attr);
 }
 
@@ -318,13 +426,12 @@ static inline int
 fi_mr_map_raw(struct fid_domain *domain, uint64_t base_addr,
 	      uint8_t *raw_key, size_t key_size, uint64_t *key, uint64_t flags)
 {
-	struct fi_mr_map_raw map = {
-		.flags = flags,
-		.base_addr = base_addr,
-		.raw_key = raw_key,
-		.key_size = key_size,
-		.key = key
-	};
+	struct fi_mr_map_raw map;
+	map.flags = flags;
+	map.base_addr = base_addr;
+	map.raw_key = raw_key;
+	map.key_size = key_size;
+	map.key = key;
 	return domain->fid.ops->control(&domain->fid, FI_MAP_RAW_MR, &map);
 }
 
@@ -343,7 +450,8 @@ static inline int
 fi_mr_refresh(struct fid_mr *mr, const struct iovec *iov, size_t count,
 	      uint64_t flags)
 {
-	struct fi_mr_modify modify = {0};
+	struct fi_mr_modify modify;
+	memset(&modify, 0, sizeof(modify));
 	modify.flags = flags;
 	modify.attr.mr_iov = iov;
 	modify.attr.iov_count = count;
