@@ -60,6 +60,25 @@ enum {
 enum {
 	FI_OPT_MIN_MULTI_RECV,		/* size_t */
 	FI_OPT_CM_DATA_SIZE,		/* size_t */
+	FI_OPT_BUFFERED_MIN,		/* size_t */
+	FI_OPT_BUFFERED_LIMIT,		/* size_t */
+	FI_OPT_SEND_BUF_SIZE,
+	FI_OPT_RECV_BUF_SIZE,
+	FI_OPT_TX_SIZE,
+	FI_OPT_RX_SIZE,
+	FI_OPT_FI_HMEM_P2P,		/* int */
+	FI_OPT_XPU_TRIGGER,		/* struct fi_trigger_xpu */
+};
+
+/*
+ * Parameters for FI_OPT_HMEM_P2P to allow endpoint control over peer to peer
+ * support and FI_HMEM.
+ */
+enum {
+	FI_HMEM_P2P_ENABLED,	/* Provider decides when to use P2P, default. */
+	FI_HMEM_P2P_REQUIRED,	/* Must use P2P for all transfers */
+	FI_HMEM_P2P_PREFERRED,	/* Should use P2P for all transfers if available */
+	FI_HMEM_P2P_DISABLED	/* Do not use P2P */
 };
 
 struct fi_ops_ep {
@@ -105,6 +124,7 @@ struct fi_ops_cm;
 struct fi_ops_rma;
 struct fi_ops_tagged;
 struct fi_ops_atomic;
+struct fi_ops_collective;
 
 /*
  * Calls which modify the properties of a endpoint (control, setopt, bind, ...)
@@ -123,6 +143,7 @@ struct fid_ep {
 	struct fi_ops_rma	*rma;
 	struct fi_ops_tagged	*tagged;
 	struct fi_ops_atomic	*atomic;
+	struct fi_ops_collective *collective;
 };
 
 struct fid_pep {
@@ -156,6 +177,18 @@ fi_endpoint(struct fid_domain *domain, struct fi_info *info,
 	return domain->ops->endpoint(domain, info, ep, context);
 }
 
+static inline int
+fi_endpoint2(struct fid_domain *domain, struct fi_info *info,
+	     struct fid_ep **ep, uint64_t flags, void *context)
+{
+	if (!flags)
+		return fi_endpoint(domain, info, ep, context);
+
+	return FI_CHECK_OP(domain->ops, struct fi_ops_domain, endpoint2) ?
+		domain->ops->endpoint2(domain, info, ep, flags, context) :
+		-FI_ENOSYS;
+}
+
 static inline int
 fi_scalable_ep(struct fid_domain *domain, struct fi_info *info,
 	    struct fid_ep **sep, void *context)
@@ -217,17 +250,17 @@ static inline int fi_ep_alias(struct fid_ep *ep, struct fid_ep **alias_ep,
 }
 
 static inline int
-fi_tx_context(struct fid_ep *ep, int index, struct fi_tx_attr *attr,
+fi_tx_context(struct fid_ep *ep, int idx, struct fi_tx_attr *attr,
 	      struct fid_ep **tx_ep, void *context)
 {
-	return ep->ops->tx_ctx(ep, index, attr, tx_ep, context);
+	return ep->ops->tx_ctx(ep, idx, attr, tx_ep, context);
 }
 
 static inline int
-fi_rx_context(struct fid_ep *ep, int index, struct fi_rx_attr *attr,
+fi_rx_context(struct fid_ep *ep, int idx, struct fi_rx_attr *attr,
 	      struct fid_ep **rx_ep, void *context)
 {
-	return ep->ops->rx_ctx(ep, index, attr, rx_ep, context);
+	return ep->ops->rx_ctx(ep, idx, attr, rx_ep, context);
 }
 
 static inline FI_DEPRECATED_FUNC ssize_t
