@@ -2,6 +2,7 @@
  * Copyright (c) 2016 Intel Corporation.  All rights reserved.
  * Copyright (c) 2016 Cisco Systems, Inc.  All rights reserved.
  * Copyright (c) 2018 Amazon.com, Inc. or its affiliates. All rights reserved.
+ * Copyright (c) 2022 DataDirect Networks, Inc. All rights reserved.
  *
  * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
@@ -226,6 +227,10 @@ extern "C" {
 	 ((err) == EWOULDBLOCK)		||	\
 	 ((err) == EAGAIN))
 
+#define OFI_SOCK_TRY_ACCEPT_AGAIN(err)		\
+	(((err) == EAGAIN)		||	\
+	 ((err) == EWOULDBLOCK))
+
 #define OFI_SOCK_TRY_CONN_AGAIN(err)		\
 	(((err) == EWOULDBLOCK)		||	\
 	 ((err) == EINPROGRESS))
@@ -273,6 +278,7 @@ do						\
 
 #define htonll _byteswap_uint64
 #define ntohll _byteswap_uint64
+#define be64toh ntohll
 #define strncasecmp _strnicmp
 
 typedef int pid_t;
@@ -678,6 +684,16 @@ static inline char* strsep(char **stringp, const char *delim)
 
 #define __attribute__(x)
 
+static inline int ofi_mmap_anon_pages(void **memptr, size_t size, int flags)
+{
+	return -FI_ENOSYS;
+}
+
+static inline int ofi_unmap_anon_pages(void *memptr, size_t size)
+{
+	return -FI_ENOSYS;
+}
+
 static inline int ofi_memalign(void **memptr, size_t alignment, size_t size)
 {
 	*memptr = _aligned_malloc(size, alignment);
@@ -759,6 +775,10 @@ ofi_sendto_socket(SOCKET fd, const void *buf, size_t count, int flags,
 	return sendto(fd, (const char*) buf, len, flags, to, (int) tolen);
 }
 
+ssize_t ofi_sendv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt,
+			 int flags);
+ssize_t ofi_recvv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt,
+			 int flags);
 ssize_t ofi_writev_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt);
 ssize_t ofi_readv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt);
 ssize_t ofi_sendmsg_tcp(SOCKET fd, const struct msghdr *msg, int flags);
@@ -770,7 +790,7 @@ ofi_sendmsg_udp(SOCKET fd, const struct msghdr *msg, int flags)
 	DWORD bytes;
 	int ret;
 
-	ret = WSASendMsg(fd, msg, flags, &bytes, NULL, NULL);
+	ret = WSASendMsg(fd, (LPWSAMSG)msg, flags, &bytes, NULL, NULL);
 	return ret ? ret : bytes;
 }
 
@@ -792,6 +812,12 @@ static inline int fi_fd_nonblock(SOCKET fd)
 	return ioctlsocket(fd, FIONBIO, &argp) ? -WSAGetLastError() : 0;
 }
 
+static inline int fi_fd_block(SOCKET fd)
+{
+	u_long argp = 0;
+	return ioctlsocket(fd, FIONBIO, &argp) ? -WSAGetLastError() : 0;
+}
+
 /* Note: Use static variable `errno` for libc routines
  * (such as fopen, lseek and etc)
  * If you need to define which function/variable is needed
@@ -814,7 +840,7 @@ static inline int ofi_syserr(void)
 	return winerr2bsderr(GetLastError());
 }
 
-static inline int fi_wait_cond(pthread_cond_t *cond, pthread_mutex_t *mut, int timeout_ms)
+static inline int ofi_wait_cond(pthread_cond_t *cond, pthread_mutex_t *mut, int timeout_ms)
 {
 	return !SleepConditionVariableCS(cond, mut, (DWORD)timeout_ms);
 }
@@ -871,7 +897,7 @@ static inline long ofi_sysconf(int name)
 		return si.dwNumberOfProcessors;
 	case _SC_PHYS_PAGES:
 		GetPhysicallyInstalledSystemMemory(&mem_size);
-		return mem_size / si.dwPageSize;
+		return (long)(mem_size / si.dwPageSize);
 	default:
 		errno = EINVAL;
 		return -1;
@@ -890,11 +916,6 @@ static inline int ofi_alloc_hugepage_buf(void **memptr, size_t size)
 	return -FI_ENOSYS;
 }
 
-static inline int ofi_free_hugepage_buf(void *memptr, size_t size)
-{
-	return -FI_ENOSYS;
-}
-
 static inline int ofi_hugepage_enabled(void)
 {
 	return 0;
@@ -1005,10 +1026,10 @@ OFI_DEF_COMPLEX(long_double)
 typedef LONG ofi_atomic_int_32_t;
 typedef LONGLONG ofi_atomic_int_64_t;
 
-#define ofi_atomic_add_and_fetch(radix, ptr, val) InterlockedAdd##radix((ofi_atomic_int_##radix##_t *)(ptr), (ofi_atomic_int_##radix##_t)(val))
-#define ofi_atomic_sub_and_fetch(radix, ptr, val) InterlockedAdd##radix((ofi_atomic_int_##radix##_t *)(ptr), -(ofi_atomic_int_##radix##_t)(val))
+#define ofi_atomic_add_and_fetch(radix, ptr, val) InterlockedAdd##radix((ofi_atomic_int_##radix##_t volatile *)(ptr), (ofi_atomic_int_##radix##_t)(val))
+#define ofi_atomic_sub_and_fetch(radix, ptr, val) InterlockedAdd##radix((ofi_atomic_int_##radix##_t volatile *)(ptr), -(ofi_atomic_int_##radix##_t)(val))
 #define ofi_atomic_cas_bool(radix, ptr, expected, desired)					\
-	(InterlockedCompareExchange##radix(ptr, desired, expected) == expected)
+	(InterlockedCompareExchange##radix((ofi_atomic_int_##radix##_t volatile *)ptr, desired, expected) == expected)
 
 #endif /* HAVE_BUILTIN_ATOMICS */
 
@@ -1026,12 +1047,12 @@ static inline int ofi_set_thread_affinity(const char *s)
 static inline void
 ofi_cpuid(unsigned func, unsigned subfunc, unsigned cpuinfo[4])
 {
-	__cpuidex(cpuinfo, func, subfunc);
+	__cpuidex((int *)cpuinfo, func, subfunc);
 }
 
-#define ofi_clwb(addr) do { _mm_clflush(addr); _mm_sfence(); } while (0)
-#define ofi_clflushopt(addr) do { _mm_clflush(addr); _mm_sfence(); } while (0)
-#define ofi_clflush(addr) _mm_clflush(addr)
+#define ofi_clwb(addr) do { _mm_clflush((void const *)addr); _mm_sfence(); } while (0)
+#define ofi_clflushopt(addr) do { _mm_clflush((void const *)addr); _mm_sfence(); } while (0)
+#define ofi_clflush(addr) _mm_clflush((void const *)addr)
 #define ofi_sfence() _mm_sfence()
 
 #else /* defined(_M_X64) || defined(_M_AMD64) */
