@@ -36,8 +36,8 @@
 #include <stdint.h>
 #include <pthread.h>
 
-#include <fi.h>
-#include <fi_list.h>
+#include <ofi.h>
+#include <ofi_list.h>
 
 #include "gnix_util.h"
 
