@@ -44,8 +44,8 @@ static int rxd_tree_compare(struct ofi_rbmap *map, void *key, void *data)
 
 	memset(addr, 0, len);
 	av = container_of(map, struct rxd_av, rbmap);
-	dg_addr = (intptr_t)ofi_idx_lookup(&av->rxdaddr_dg_idx, 
-					   (fi_addr_t) data);
+	dg_addr = (intptr_t)ofi_idx_lookup(&av->rxdaddr_dg_idx,
+					   (int)(uintptr_t) data);
 
 	ret = fi_av_lookup(av->dg_av, dg_addr,addr, &len);
 	if (ret)
@@ -109,11 +109,11 @@ close:
 static fi_addr_t rxd_av_dg_addr(struct rxd_av *av, fi_addr_t fi_addr)
 {
 	fi_addr_t dg_addr;
-	fi_addr_t rxd_addr = (intptr_t) ofi_idx_lookup(&av->fi_addr_idx, 
-					     RXD_IDX_OFFSET(fi_addr));
+	fi_addr_t rxd_addr = (intptr_t) ofi_idx_lookup(&av->fi_addr_idx,
+					     RXD_IDX_OFFSET((int)fi_addr));
 	if (!rxd_addr)
 		goto err;
-	dg_addr = (intptr_t) ofi_idx_lookup(&av->rxdaddr_dg_idx, rxd_addr);
+	dg_addr = (intptr_t) ofi_idx_lookup(&av->rxdaddr_dg_idx, (int)rxd_addr);
 	if (!dg_addr)
 		goto err;
 
@@ -133,25 +133,26 @@ static int rxd_set_rxd_addr(struct rxd_av *av, fi_addr_t dg_addr, fi_addr_t *add
 
 }
 
-static fi_addr_t rxd_set_fi_addr(struct rxd_av *av, fi_addr_t rxd_addr)
+static int rxd_set_fi_addr(struct rxd_av *av, fi_addr_t rxd_addr)
 {
-	int fi_addr;
+	int idx;
 	fi_addr_t dg_addr;
-	fi_addr = ofi_idx_insert(&(av->fi_addr_idx), (void*)(uintptr_t)rxd_addr);
-	if (fi_addr < 0)
+
+	idx = ofi_idx_insert(&(av->fi_addr_idx), (void*)(uintptr_t)rxd_addr);
+	if (idx < 0)
 		goto nomem1;
-	
-	if (ofi_idm_set(&(av->rxdaddr_fi_idm), rxd_addr, 
-		        (void*)(uintptr_t) fi_addr) < 0)
+
+	if (ofi_idm_set(&(av->rxdaddr_fi_idm), (int)rxd_addr,
+		        (void*)(uintptr_t) idx) < 0)
 		goto nomem2;
 
-	return fi_addr;
+	return idx;
 
 nomem2:
-	ofi_idx_remove_ordered(&(av->fi_addr_idx), fi_addr);
+	ofi_idx_remove_ordered(&(av->fi_addr_idx), idx);
 nomem1:
-	dg_addr = (intptr_t) ofi_idx_remove_ordered(&(av->rxdaddr_dg_idx), 
-						    rxd_addr);
+	dg_addr = (intptr_t) ofi_idx_remove_ordered(&(av->rxdaddr_dg_idx),
+						    (int) rxd_addr);
 	fi_av_remove(av->dg_av, &dg_addr, 1, 0);
 
 	return -FI_ENOMEM;
@@ -178,7 +179,7 @@ int rxd_av_insert_dg_addr(struct rxd_av *av, const void *addr,
 			       NULL);
 	if (ret) {
 		assert(ret != -FI_EALREADY);
-		ofi_idx_remove_ordered(&(av->rxdaddr_dg_idx), *rxd_addr);
+		ofi_idx_remove_ordered(&(av->rxdaddr_dg_idx), (int)(*rxd_addr));
 		goto nomem;
 	}
 
@@ -195,11 +196,20 @@ static int rxd_av_insert(struct fid_av *av_fid, const void *addr, size_t count,
 	struct rxd_av *av;
 	int i = 0, ret = 0, success_cnt = 0;
 	fi_addr_t rxd_addr;
-	int util_addr;
+	int util_addr, *sync_err = NULL;
 	struct ofi_rbnode *node;
 
 	av = container_of(av_fid, struct rxd_av, util_av.av_fid);
-	fastlock_acquire(&av->util_av.lock);
+	ret = ofi_verify_av_insert(&av->util_av, flags, context);
+	if (ret)
+		return ret;
+
+	if (flags & FI_SYNC_ERR) {
+		sync_err = context;
+		memset(sync_err, 0, sizeof(*sync_err) * count);
+	}
+
+	ofi_mutex_lock(&av->util_av.lock);
 	if (!av->dg_addrlen) {
 		ret = rxd_av_set_addrlen(av, addr);
 		if (ret)
@@ -212,19 +222,19 @@ static int rxd_av_insert(struct fid_av *av_fid, const void *addr, size_t count,
 			rxd_addr = (fi_addr_t) node->data;
 		} else {
 			ret = rxd_av_insert_dg_addr(av, addr, &rxd_addr,
-						    flags, context);
+						    flags, sync_err ?
+						    &sync_err[i] : context);
 			if (ret)
 				break;
 		}
 
-		util_addr = (intptr_t)ofi_idm_lookup(&av->rxdaddr_fi_idm, 
-						     rxd_addr);
-
+		util_addr = (int)(intptr_t) ofi_idm_lookup(&av->rxdaddr_fi_idm,
+							   (int) rxd_addr);
 		if (!util_addr) {
 			util_addr = rxd_set_fi_addr(av, rxd_addr);
 			if (util_addr < 0) {
 				ret = util_addr;
-				break;	
+				break;
 			}
 		}
 		if (fi_addr)
@@ -237,21 +247,25 @@ static int rxd_av_insert(struct fid_av *av_fid, const void *addr, size_t count,
 		FI_WARN(&rxd_prov, FI_LOG_AV,
 			"failed to insert address %d: %d (%s)\n",
 			i, -ret, fi_strerror(-ret));
-		if (av->util_av.eq)
-			ofi_av_write_event(&av->util_av, i, -ret, context);
 		if (fi_addr)
 			fi_addr[i] = FI_ADDR_NOTAVAIL;
+		if (av->util_av.eq)
+			ofi_av_write_event(&av->util_av, i, -ret, context);
+		else if (sync_err)
+			sync_err[i] = -ret;
 		i++;
 	}
 out:
 	av->dg_av_used += success_cnt;
-	fastlock_release(&av->util_av.lock);
+	ofi_mutex_unlock(&av->util_av.lock);
 
 	for (; i < count; i++) {
-		if (av->util_av.eq)
-			ofi_av_write_event(&av->util_av, i, FI_ECANCELED, context);
 		if (fi_addr)
 			fi_addr[i] = FI_ADDR_NOTAVAIL;
+		if (av->util_av.eq)
+			ofi_av_write_event(&av->util_av, i, FI_ECANCELED, context);
+		else if (sync_err)
+			sync_err[i] = FI_ECANCELED;
 	}
 
 	if (av->util_av.eq) {
@@ -280,49 +294,28 @@ static int rxd_av_remove(struct fid_av *av_fid, fi_addr_t *fi_addr, size_t count
 			uint64_t flags)
 {
 	int ret = 0;
-	size_t i, addrlen;
+	size_t i;
 	fi_addr_t rxd_addr;
-	fi_addr_t dg_addr;
 	struct rxd_av *av;
-	uint8_t addr[RXD_NAME_LENGTH];
 
 	av = container_of(av_fid, struct rxd_av, util_av.av_fid);
-	fastlock_acquire(&av->util_av.lock);
+	ofi_mutex_lock(&av->util_av.lock);
 	for (i = 0; i < count; i++) {
-
-		addrlen = RXD_NAME_LENGTH;
 		rxd_addr = (intptr_t)ofi_idx_lookup(&av->fi_addr_idx,
-						    RXD_IDX_OFFSET(fi_addr[i]));
+						    (int) RXD_IDX_OFFSET(fi_addr[i]));
 		if (!rxd_addr)
 			goto err;
-		
-		dg_addr = (intptr_t)ofi_idx_lookup(&av->rxdaddr_dg_idx, rxd_addr);
 
-		ret = fi_av_lookup(av->dg_av, dg_addr, addr, &addrlen);
-		if (ret)
-			goto err;
-		
-		ret = ofi_rbmap_find_delete(&av->rbmap, (void *) addr);
-		if (ret)
-			goto err;
-
-		ret = fi_av_remove(av->dg_av, &dg_addr, 1, flags);
-
-		if (ret)
-			goto err;
-		
-		ofi_idx_remove_ordered(&(av->fi_addr_idx), 
-				       RXD_IDX_OFFSET(fi_addr[i]));
-		ofi_idx_remove_ordered(&(av->rxdaddr_dg_idx), rxd_addr);
-		ofi_idm_clear(&(av->rxdaddr_fi_idm), rxd_addr);
-		av->dg_av_used--;
+		ofi_idx_remove_ordered(&(av->fi_addr_idx),
+				       (int) RXD_IDX_OFFSET(fi_addr[i]));
+		ofi_idm_clear(&(av->rxdaddr_fi_idm), (int) rxd_addr);
 	}
 
 err:
 	if (ret)
 		FI_WARN(&rxd_prov, FI_LOG_AV, "Unable to remove address from AV\n");
 
-	fastlock_release(&av->util_av.lock);
+	ofi_mutex_unlock(&av->util_av.lock);
 	return ret;
 }
 
@@ -361,22 +354,39 @@ static struct fi_ops_av rxd_av_ops = {
 static int rxd_av_close(struct fid *fid)
 {
 	struct rxd_av *av;
+	struct ofi_rbnode *node;
+	fi_addr_t dg_addr, rxd_addr;
 	int ret;
 
-
 	av = container_of(fid, struct rxd_av, util_av.av_fid);
-	ret = fi_close(&av->dg_av->fid);
-	if (ret)
-		return ret;
 
-	ofi_rbmap_cleanup(&av->rbmap);
 	ret = ofi_av_close(&av->util_av);
 	if (ret)
 		return ret;
 
-	ofi_idx_reset(&(av->fi_addr_idx)); 
+	while ((node = ofi_rbmap_get_root(&av->rbmap))) {
+		rxd_addr = (fi_addr_t) node->data;
+		dg_addr = (intptr_t)ofi_idx_lookup(&av->rxdaddr_dg_idx,
+						   (int) rxd_addr);
+
+		ret = fi_av_remove(av->dg_av, &dg_addr, 1, 0);
+		if (ret)
+			FI_WARN(&rxd_prov, FI_LOG_AV,
+				"failed to remove dg addr: %d (%s)\n",
+				-ret, fi_strerror(-ret));
+
+		ofi_idx_remove_ordered(&(av->rxdaddr_dg_idx), (int) rxd_addr);
+		ofi_rbmap_delete(&av->rbmap, node);
+	}
+	ofi_rbmap_cleanup(&av->rbmap);
+
+	ret = fi_close(&av->dg_av->fid);
+	if (ret)
+		return ret;
+
+	ofi_idx_reset(&(av->fi_addr_idx));
 	ofi_idx_reset(&(av->rxdaddr_dg_idx));
-	ofi_idm_reset(&(av->rxdaddr_fi_idm));
+	ofi_idm_reset(&(av->rxdaddr_fi_idm), NULL);
 
 	free(av);
 	return 0;
@@ -433,7 +443,7 @@ int rxd_av_create(struct fid_domain *domain_fid, struct fi_av_attr *attr,
 		goto err1;
 
 	ofi_rbmap_init(&av->rbmap, rxd_tree_compare);
-	
+
 	av_attr = *attr;
 	av_attr.count = 0;
 	av_attr.flags = 0;
@@ -448,7 +458,7 @@ int rxd_av_create(struct fid_domain *domain_fid, struct fi_av_attr *attr,
 
 err2:
 	ofi_av_close(&av->util_av);
-err1:	
+err1:
 	free(av);
 	return ret;
 }
