@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2016 Intel Corporation, Inc.  All rights reserved.
+ * (C) Copyright 2020 Hewlett Packard Enterprise Development LP
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -38,24 +39,176 @@
 #include <ofi_coll.h>
 #include "rxm.h"
 
-int rxm_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
-		  struct fid_cntr **cntr_fid, void *context)
-{
-	int ret;
-	struct util_cntr *cntr;
 
+static uint64_t rxm_passthru_cntr_read(struct fid_cntr *cntr_fid)
+{
+	struct rxm_cntr *cntr;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	cntr->util_cntr.progress(&cntr->util_cntr);
+	return fi_cntr_read(cntr->msg_cntr);
+}
+
+static uint64_t rxm_passthru_cntr_readerr(struct fid_cntr *cntr_fid)
+{
+	struct rxm_cntr *cntr;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	cntr->util_cntr.progress(&cntr->util_cntr);
+	return fi_cntr_readerr(cntr->msg_cntr);
+}
+
+static int rxm_passthru_cntr_add(struct fid_cntr *cntr_fid, uint64_t value)
+{
+	struct rxm_cntr *cntr;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	return fi_cntr_add(cntr->msg_cntr, value);
+}
+
+static int rxm_passthru_cntr_set(struct fid_cntr *cntr_fid, uint64_t value)
+{
+	struct rxm_cntr *cntr;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	return fi_cntr_set(cntr->msg_cntr, value);
+}
+
+
+static int rxm_passthru_cntr_wait(struct fid_cntr *cntr_fid,
+				  uint64_t threshold, int timeout)
+{
+	struct rxm_cntr *cntr;
+	uint64_t endtime, errcnt;
+	int ret, timeout_quantum;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	errcnt = fi_cntr_readerr(cntr->msg_cntr);
+	endtime = ofi_timeout_time(timeout);
+
+	do {
+		cntr->util_cntr.progress(&cntr->util_cntr);
+
+		if (errcnt != fi_cntr_readerr(cntr->msg_cntr))
+			return -FI_EAVAIL;
+
+		if (ofi_adjust_timeout(endtime, &timeout))
+			return -FI_ETIMEDOUT;
+
+		/* We need to ensure that CM events are progressed, so return
+		 * periodically to force progress on the EQ.
+		 */
+		timeout_quantum = (timeout < 0 ? OFI_TIMEOUT_QUANTUM_MS :
+				   MIN(OFI_TIMEOUT_QUANTUM_MS, timeout));
+
+		ret = fi_cntr_wait(cntr->msg_cntr, threshold, timeout_quantum);
+	} while (ret == -FI_ETIMEDOUT &&
+	         (timeout < 0 || timeout_quantum < timeout));
+
+	return ret;
+}
+
+static int rxm_passthru_cntr_adderr(struct fid_cntr *cntr_fid, uint64_t value)
+{
+	struct rxm_cntr *cntr;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	return fi_cntr_adderr(cntr->msg_cntr, value);
+}
+
+static int rxm_passthru_cntr_seterr(struct fid_cntr *cntr_fid, uint64_t value)
+{
+	struct rxm_cntr *cntr;
+
+	cntr = container_of(cntr_fid, struct rxm_cntr, util_cntr.cntr_fid);
+	return fi_cntr_seterr(cntr->msg_cntr, value);
+}
+
+static struct fi_ops_cntr rxm_passthru_cntr_ops = {
+	.size = sizeof(struct fi_ops_cntr),
+	.read = rxm_passthru_cntr_read,
+	.readerr = rxm_passthru_cntr_readerr,
+	.add = rxm_passthru_cntr_add,
+	.set = rxm_passthru_cntr_set,
+	.wait = rxm_passthru_cntr_wait,
+	.adderr = rxm_passthru_cntr_adderr,
+	.seterr = rxm_passthru_cntr_seterr,
+};
+
+static int rxm_passthru_cntr_close(struct fid *fid)
+{
+	struct rxm_cntr *cntr;
+	int ret;
+
+	cntr = container_of(fid, struct rxm_cntr, util_cntr.cntr_fid.fid);
+
+	if (cntr->msg_cntr) {
+		ret = fi_close(&cntr->msg_cntr->fid);
+		if (ret)
+			return ret;
+	}
+
+	cntr->msg_cntr = NULL;
+	ret = ofi_cntr_cleanup(&cntr->util_cntr);
+	if (ret)
+		return ret;
+
+	free(cntr);
+	return 0;
+}
+
+static struct fi_ops rxm_passthru_cntr_fi_ops = {
+	.size = sizeof(struct fi_ops),
+	.close = rxm_passthru_cntr_close,
+	.bind = fi_no_bind,
+	.control = fi_no_control,
+	.ops_open = fi_no_ops_open,
+};
+
+static int
+rxm_av_open(struct fid_domain *domain_fid, struct fi_av_attr *attr,
+	    struct fid_av **fid_av, void *context)
+{
+	return rxm_util_av_open(domain_fid, attr, fid_av,
+			context, sizeof(struct rxm_conn),
+			ofi_av_remove_cleanup ? rxm_av_remove_handler : NULL);
+}
+
+static int
+rxm_cntr_open(struct fid_domain *fid_domain, struct fi_cntr_attr *attr,
+	      struct fid_cntr **cntr_fid, void *context)
+{
+	struct rxm_domain *domain;
+	struct rxm_cntr *cntr;
+	int ret;
+
+	domain = container_of(fid_domain, struct rxm_domain,
+			      util_domain.domain_fid.fid);
 	cntr = calloc(1, sizeof(*cntr));
 	if (!cntr)
 		return -FI_ENOMEM;
 
-	ret = ofi_cntr_init(&rxm_prov, domain, attr, cntr,
+	ret = ofi_cntr_init(&rxm_prov, fid_domain, attr, &cntr->util_cntr,
 			    &ofi_cntr_progress, context);
 	if (ret)
 		goto free;
 
-	*cntr_fid = &cntr->cntr_fid;
+	if (domain->passthru) {
+		ret = fi_cntr_open(domain->msg_domain, attr, &cntr->msg_cntr, cntr);
+		if (ret) {
+			RXM_WARN_ERR(FI_LOG_CNTR, "fi_cntr_open", ret);
+			goto cleanup;
+		}
+
+		cntr->util_cntr.cntr_fid.fid.ops = &rxm_passthru_cntr_fi_ops;
+		cntr->util_cntr.cntr_fid.ops = &rxm_passthru_cntr_ops;
+	}
+
+	*cntr_fid = &cntr->util_cntr.cntr_fid;
 	return FI_SUCCESS;
 
+cleanup:
+	ofi_cntr_cleanup(&cntr->util_cntr);
 free:
 	free(cntr);
 	return ret;
@@ -77,10 +230,10 @@ static struct fi_ops_domain rxm_domain_ops = {
 
 static void rxm_mr_remove_map_entry(struct rxm_mr *mr)
 {
-	fastlock_acquire(&mr->domain->util_domain.lock);
+	ofi_genlock_lock(&mr->domain->util_domain.lock);
 	(void) ofi_mr_map_remove(&mr->domain->util_domain.mr_map,
 				 mr->mr_fid.key);
-	fastlock_release(&mr->domain->util_domain.lock);
+	ofi_genlock_unlock(&mr->domain->util_domain.lock);
 }
 
 static int rxm_mr_add_map_entry(struct util_domain *domain,
@@ -92,7 +245,7 @@ static int rxm_mr_add_map_entry(struct util_domain *domain,
 
 	msg_attr->requested_key = rxm_mr->mr_fid.key;
 
-	fastlock_acquire(&domain->lock);
+	ofi_genlock_lock(&domain->lock);
 	ret = ofi_mr_map_insert(&domain->mr_map, msg_attr, &temp_key, rxm_mr);
 	if (OFI_UNLIKELY(ret)) {
 		FI_WARN(&rxm_prov, FI_LOG_DOMAIN,
@@ -101,11 +254,22 @@ static int rxm_mr_add_map_entry(struct util_domain *domain,
 	} else {
 		assert(rxm_mr->mr_fid.key == temp_key);
 	}
-	fastlock_release(&domain->lock);
+	ofi_genlock_unlock(&domain->lock);
 
 	return ret;
 }
 
+struct rxm_mr *rxm_mr_get_map_entry(struct rxm_domain *domain, uint64_t key)
+{
+	struct rxm_mr *mr;
+
+	ofi_genlock_lock(&domain->util_domain.lock);
+	mr = ofi_mr_map_get(&domain->util_domain.mr_map, key);
+	ofi_genlock_unlock(&domain->util_domain.lock);
+
+	return mr;
+}
+
 static int rxm_domain_close(fid_t fid)
 {
 	struct rxm_domain *rxm_domain;
@@ -113,6 +277,9 @@ static int rxm_domain_close(fid_t fid)
 
 	rxm_domain = container_of(fid, struct rxm_domain, util_domain.domain_fid.fid);
 
+	ofi_mutex_destroy(&rxm_domain->amo_bufpool_lock);
+	ofi_bufpool_destroy(rxm_domain->amo_bufpool);
+
 	ret = fi_close(&rxm_domain->msg_domain->fid);
 	if (ret)
 		return ret;
@@ -216,14 +383,18 @@ err:
 	return ret;
 }
 
+/* Large send/recv transfers use RMA rendezvous protocol */
 static uint64_t
 rxm_mr_get_msg_access(struct rxm_domain *rxm_domain, uint64_t access)
 {
-	/* Additional flags to use RMA read for large message transfers */
-	access |= FI_READ | FI_REMOTE_READ;
+	if (access & FI_SEND) {
+		access |= rxm_use_write_rndv ? FI_WRITE : FI_REMOTE_READ;
+	}
+
+	if (access & FI_RECV) {
+		access |= rxm_use_write_rndv ? FI_REMOTE_WRITE : FI_READ;
+	}
 
-	if (rxm_domain->mr_local)
-		access |= FI_WRITE;
 	return access;
 }
 
@@ -233,10 +404,7 @@ static void rxm_mr_init(struct rxm_mr *rxm_mr, struct rxm_domain *domain,
 	rxm_mr->mr_fid.fid.fclass = FI_CLASS_MR;
 	rxm_mr->mr_fid.fid.context = context;
 	rxm_mr->mr_fid.fid.ops = &rxm_mr_ops;
-	/* Store msg_mr as rxm_mr descriptor so that we can get its key when
-	 * the app passes msg_mr as the descriptor in fi_send and friends.
-	 * The key would be used in large message transfer protocol and RMA. */
-	rxm_mr->mr_fid.mem_desc = rxm_mr->msg_mr;
+	rxm_mr->mr_fid.mem_desc = rxm_mr;
 	rxm_mr->mr_fid.key = fi_mr_key(rxm_mr->msg_mr);
 	rxm_mr->domain = domain;
 	ofi_atomic_inc32(&domain->util_domain.ref);
@@ -253,10 +421,23 @@ static int rxm_mr_regattr(struct fid *fid, const struct fi_mr_attr *attr,
 	rxm_domain = container_of(fid, struct rxm_domain,
 				  util_domain.domain_fid.fid);
 
+	if (!ofi_hmem_is_initialized(attr->iface)) {
+		FI_WARN(&rxm_prov, FI_LOG_MR,
+			"Cannot register memory for uninitialized iface\n");
+		return -FI_ENOSYS;
+	}
+
 	rxm_mr = calloc(1, sizeof(*rxm_mr));
 	if (!rxm_mr)
 		return -FI_ENOMEM;
 
+	ofi_mr_update_attr(rxm_domain->util_domain.fabric->fabric_fid.api_version,
+			   rxm_domain->util_domain.info_domain_caps, attr,
+			   &msg_attr);
+
+	if ((flags & FI_HMEM_HOST_ALLOC) && (attr->iface == FI_HMEM_ZE))
+		msg_attr.device.ze = -1;
+
 	msg_attr.access = rxm_mr_get_msg_access(rxm_domain, attr->access);
 
 	ret = fi_mr_regattr(rxm_domain->msg_domain, &msg_attr,
@@ -266,6 +447,9 @@ static int rxm_mr_regattr(struct fid *fid, const struct fi_mr_attr *attr,
 		goto err;
 	}
 	rxm_mr_init(rxm_mr, rxm_domain, attr->context);
+	ofi_mutex_init(&rxm_mr->amo_lock);
+	rxm_mr->iface = msg_attr.iface;
+	rxm_mr->device = msg_attr.device.reserved;
 	*mr = &rxm_mr->mr_fid;
 
 	if (rxm_domain->util_domain.info_domain_caps & FI_ATOMIC) {
@@ -355,31 +539,74 @@ static struct fi_ops_mr rxm_domain_mr_ops = {
 	.regattr = rxm_mr_regattr,
 };
 
+static int
+rxm_mr_regattr_thru(struct fid *fid, const struct fi_mr_attr *attr,
+		    uint64_t flags, struct fid_mr **mr)
+{
+	struct rxm_domain *domain;
+
+	domain = container_of(fid, struct rxm_domain,
+			      util_domain.domain_fid.fid);
+	return fi_mr_regattr(domain->msg_domain, attr, flags, mr);
+}
+
+static int
+rxm_mr_regv_thru(struct fid *fid, const struct iovec *iov, size_t count,
+		 uint64_t access, uint64_t offset, uint64_t requested_key,
+		 uint64_t flags, struct fid_mr **mr, void *context)
+{
+	struct rxm_domain *domain;
+
+	domain = container_of(fid, struct rxm_domain,
+			      util_domain.domain_fid.fid);
+	return fi_mr_regv(domain->msg_domain, iov, count, access, offset,
+			 requested_key, flags, mr, context);
+}
+
+static int rxm_mr_reg_thru(struct fid *fid, const void *buf, size_t len,
+		      uint64_t access, uint64_t offset, uint64_t requested_key,
+		      uint64_t flags, struct fid_mr **mr, void *context)
+{
+	struct rxm_domain *domain;
+
+	domain = container_of(fid, struct rxm_domain,
+			      util_domain.domain_fid.fid);
+	return fi_mr_reg(domain->msg_domain, buf, len, access, offset,
+			 requested_key, flags, mr, context);
+}
+
+static struct fi_ops_mr rxm_domain_mr_thru_ops = {
+	.size = sizeof(struct fi_ops_mr),
+	.reg = rxm_mr_reg_thru,
+	.regv = rxm_mr_regv_thru,
+	.regattr = rxm_mr_regattr_thru,
+};
+
 static ssize_t rxm_send_credits(struct fid_ep *ep, size_t credits)
 {
-	struct rxm_conn *rxm_conn =
-		container_of(ep->fid.context, struct rxm_conn, handle);
-	struct rxm_ep *rxm_ep = rxm_conn->handle.cmap->ep;
+	struct rxm_conn *rxm_conn = ep->fid.context;
+	struct rxm_ep *rxm_ep = rxm_conn->ep;
 	struct rxm_deferred_tx_entry *def_tx_entry;
-	struct rxm_tx_base_buf *tx_buf;
+	struct rxm_tx_buf *tx_buf;
 	struct iovec iov;
 	struct fi_msg msg;
 	ssize_t ret;
 
-	tx_buf = rxm_tx_buf_alloc(rxm_ep, RXM_BUF_POOL_TX_CREDIT);
+	tx_buf = ofi_buf_alloc(rxm_ep->tx_pool);
 	if (!tx_buf) {
 		FI_WARN(&rxm_prov, FI_LOG_EP_DATA,
 			"Ran out of buffers from TX credit buffer pool.\n");
 		return -FI_ENOMEM;
 	}
 
+	tx_buf->hdr.state = RXM_CREDIT_TX;
 	rxm_ep_format_tx_buf_pkt(rxm_conn, 0, rxm_ctrl_credit, 0, 0, FI_SEND,
 				 &tx_buf->pkt);
 	tx_buf->pkt.ctrl_hdr.type = rxm_ctrl_credit;
 	tx_buf->pkt.ctrl_hdr.msg_id = ofi_buf_index(tx_buf);
 	tx_buf->pkt.ctrl_hdr.ctrl_data = credits;
 
-	if (rxm_conn->handle.state != RXM_CMAP_CONNECTED)
+	if (rxm_conn->state != RXM_CM_CONNECTED)
 		goto defer;
 
 	iov.iov_base = &tx_buf->pkt;
@@ -404,13 +631,10 @@ defer:
 	}
 
 	def_tx_entry->credit_msg.tx_buf = tx_buf;
-	rxm_ep_enqueue_deferred_tx_queue(def_tx_entry);
+	rxm_queue_deferred_tx(def_tx_entry, OFI_LIST_HEAD);
 	return FI_SUCCESS;
 }
 
-static void rxm_no_set_threshold(struct fid_ep *ep_fid, size_t threshold)
-{ }
-
 static void rxm_no_add_credits(struct fid_ep *ep_fid, size_t credits)
 { }
 
@@ -418,27 +642,88 @@ static void rxm_no_credit_handler(struct fid_domain *domain_fid,
 		ssize_t (*credit_handler)(struct fid_ep *ep, size_t credits))
 { }
 
-static int rxm_no_enable_flow_ctrl(struct fid_ep *ep_fid)
+static int rxm_no_enable_flow_ctrl(struct fid_ep *ep_fid, uint64_t threshold)
 {
 	return -FI_ENOSYS;
 }
 
+static bool rxm_no_flow_ctrl_available(struct fid_ep *ep_fid)
+{
+	return false;
+}
+
 struct ofi_ops_flow_ctrl rxm_no_ops_flow_ctrl = {
 	.size = sizeof(struct ofi_ops_flow_ctrl),
-	.set_threshold = rxm_no_set_threshold,
 	.add_credits = rxm_no_add_credits,
 	.enable = rxm_no_enable_flow_ctrl,
 	.set_send_handler = rxm_no_credit_handler,
+	.available = rxm_no_flow_ctrl_available,
 };
 
+static int rxm_config_flow_ctrl(struct rxm_domain *domain)
+{
+	struct ofi_ops_flow_ctrl *flow_ctrl_ops;
+	int ret;
+
+	ret = fi_open_ops(&domain->msg_domain->fid, OFI_OPS_FLOW_CTRL, 0,
+			  (void **) &flow_ctrl_ops, NULL);
+	if (ret) {
+		if (ret == -FI_ENOSYS) {
+			domain->flow_ctrl_ops = &rxm_no_ops_flow_ctrl;
+			return 0;
+		}
+		return ret;
+	}
+
+	assert(flow_ctrl_ops);
+	domain->flow_ctrl_ops = flow_ctrl_ops;
+	domain->flow_ctrl_ops->set_send_handler(domain->msg_domain,
+						rxm_send_credits);
+	return 0;
+}
+
+struct ofi_ops_dynamic_rbuf rxm_dynamic_rbuf = {
+	.size = sizeof(struct ofi_ops_dynamic_rbuf),
+	.get_rbuf = rxm_get_dyn_rbuf,
+};
+
+static void rxm_config_dyn_rbuf(struct rxm_domain *domain, struct fi_info *info,
+				struct fi_info *msg_info)
+{
+	int ret = 1;
+
+	/* Collective support requires rxm generated and consumed messages.
+	 * Although we could update the code to handle receiving collective
+	 * messages, collective support is mostly for development purposes.
+	 * So, fallback to bounce buffers when enabled.
+	 * We also can't pass through HMEM buffers, unless the lower layer
+	 * can handle them.
+	 */
+	if (domain->passthru || (info->caps & FI_COLLECTIVE) ||
+	    ((info->caps & FI_HMEM) && !(msg_info->caps & FI_HMEM)))
+		return;
+
+	fi_param_get_bool(&rxm_prov, "enable_dyn_rbuf", &ret);
+	domain->dyn_rbuf = (ret != 0);
+	if (!domain->dyn_rbuf)
+		return;
+
+	ret = fi_set_ops(&domain->msg_domain->fid, OFI_OPS_DYNAMIC_RBUF, 0,
+			 (void *) &rxm_dynamic_rbuf, NULL);
+	domain->dyn_rbuf = (ret == FI_SUCCESS);
+
+	if (domain->dyn_rbuf) {
+		domain->rx_post_size = sizeof(struct rxm_pkt);
+	}
+}
+
 int rxm_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 		struct fid_domain **domain, void *context)
 {
-	int ret;
 	struct rxm_domain *rxm_domain;
 	struct rxm_fabric *rxm_fabric;
-	struct fi_info *msg_info;
-	struct ofi_ops_flow_ctrl *flow_ctrl_ops;
+	struct fi_info *msg_info, *base_info;
+	int ret;
 
 	rxm_domain = calloc(1, sizeof(*rxm_domain));
 	if (!rxm_domain)
@@ -446,8 +731,9 @@ int rxm_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 
 	rxm_fabric = container_of(fabric, struct rxm_fabric, util_fabric.fabric_fid);
 
+	base_info = rxm_passthru_info(info) ? &rxm_thru_info : NULL;
 	ret = ofi_get_core_info(fabric->api_version, NULL, NULL, 0, &rxm_util_prov,
-				info, NULL, rxm_info_to_core, &msg_info);
+				info, base_info, rxm_info_to_core, &msg_info);
 	if (ret)
 		goto err1;
 
@@ -456,7 +742,8 @@ int rxm_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 	if (ret)
 		goto err2;
 
-	ret = ofi_domain_init(fabric, info, &rxm_domain->util_domain, context);
+	ret = ofi_domain_init(fabric, info, &rxm_domain->util_domain, context,
+			      OFI_LOCK_MUTEX);
 	if (ret) {
 		goto err3;
 	}
@@ -468,28 +755,36 @@ int rxm_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 	rxm_domain->util_domain.mr_map.mode &= ~FI_MR_PROV_KEY;
 
 	rxm_domain->max_atomic_size = rxm_ep_max_atomic_size(info);
+	rxm_domain->rx_post_size = rxm_packet_size;
+
 	*domain = &rxm_domain->util_domain.domain_fid;
 	(*domain)->fid.ops = &rxm_domain_fi_ops;
-	/* Replace MR ops set by ofi_domain_init() */
-	(*domain)->mr = &rxm_domain_mr_ops;
 	(*domain)->ops = &rxm_domain_ops;
 
-	rxm_domain->mr_local = ofi_mr_local(msg_info) && !ofi_mr_local(info);
-
-	ret = fi_open_ops(&rxm_domain->msg_domain->fid, OFI_OPS_FLOW_CTRL, 0,
-			  (void **) &flow_ctrl_ops, NULL);
-	if (!ret && flow_ctrl_ops) {
-		rxm_domain->flow_ctrl_ops = flow_ctrl_ops;
-		rxm_domain->flow_ctrl_ops->set_send_handler(
-			rxm_domain->msg_domain, rxm_send_credits);
-	} else if (ret == -FI_ENOSYS) {
-		rxm_domain->flow_ctrl_ops = &rxm_no_ops_flow_ctrl;
-	} else {
+	ret = ofi_bufpool_create(&rxm_domain->amo_bufpool,
+				 rxm_domain->max_atomic_size, 64, 0, 0, 0);
+	if (ret)
 		goto err3;
-	}
+
+	ofi_mutex_init(&rxm_domain->amo_bufpool_lock);
+
+	rxm_domain->passthru = rxm_passthru_info(info);
+	if (rxm_domain->passthru)
+		(*domain)->mr = &rxm_domain_mr_thru_ops;
+	else
+		(*domain)->mr = &rxm_domain_mr_ops;
+
+	ret = rxm_config_flow_ctrl(rxm_domain);
+	if (ret)
+		goto err4;
+
+	rxm_config_dyn_rbuf(rxm_domain, info, msg_info);
 
 	fi_freeinfo(msg_info);
 	return 0;
+err4:
+	ofi_mutex_destroy(&rxm_domain->amo_bufpool_lock);
+	ofi_bufpool_destroy(rxm_domain->amo_bufpool);
 err3:
 	fi_close(&rxm_domain->msg_domain->fid);
 err2:
