@@ -55,35 +55,31 @@ static void smr_generic_atomic_format(struct smr_cmd *cmd, uint8_t datatype,
 
 static void smr_format_inline_atomic(struct smr_cmd *cmd,
 				     enum fi_hmem_iface iface, uint64_t device,
-				     const struct iovec *iov, size_t count,
-				     const struct iovec *compv,
-				     size_t comp_count)
+				     const struct iovec *iov, size_t count)
 {
-	size_t comp_size;
-
 	cmd->msg.hdr.op_src = smr_src_inline;
 
-	switch (cmd->msg.hdr.op) {
-	case ofi_op_atomic:
-	case ofi_op_atomic_fetch:
-		cmd->msg.hdr.size = ofi_copy_from_hmem_iov(cmd->msg.data.msg,
-						SMR_MSG_DATA_LEN, iface, device,
-						iov, count, 0);
-		break;
-	case ofi_op_atomic_compare:
-		cmd->msg.hdr.size = ofi_copy_from_hmem_iov(cmd->msg.data.buf,
-						SMR_MSG_DATA_LEN, iface, device,
-						iov, count, 0);
-		comp_size = ofi_copy_from_hmem_iov(cmd->msg.data.comp,
-						SMR_MSG_DATA_LEN, iface, device,
-						compv, comp_count, 0);
-		if (comp_size != cmd->msg.hdr.size)
-			FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
-				"atomic and compare buffer size mismatch\n");
-		break;
-	default:
-		break;
-	}
+	cmd->msg.hdr.size = ofi_copy_from_hmem_iov(cmd->msg.data.msg,
+					SMR_MSG_DATA_LEN, iface, device,
+					iov, count, 0);
+}
+
+static void smr_do_atomic_inline(struct smr_ep *ep, struct smr_region *peer_smr,
+			int64_t id, int64_t peer_id, uint32_t op,
+			uint64_t op_flags, enum fi_hmem_iface iface,
+			uint64_t device, uint8_t datatype, uint8_t atomic_op,
+			const struct iovec *iov, size_t iov_count,
+			size_t total_len)
+{
+	struct smr_cmd *cmd;
+
+	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
+	smr_generic_format(cmd, peer_id, op, 0, 0, op_flags);
+	smr_generic_atomic_format(cmd, datatype, atomic_op);
+	smr_format_inline_atomic(cmd, iface, device, iov, iov_count);
+
+	ofi_cirque_commit(smr_cmd_queue(peer_smr));
+	peer_smr->cmd_cnt--;
 }
 
 static void smr_format_inject_atomic(struct smr_cmd *cmd,
@@ -100,6 +96,10 @@ static void smr_format_inject_atomic(struct smr_cmd *cmd,
 
 	switch (cmd->msg.hdr.op) {
 	case ofi_op_atomic:
+		cmd->msg.hdr.size = ofi_copy_from_hmem_iov(tx_buf->data,
+					SMR_INJECT_SIZE, iface, device,
+					iov, count, 0);
+		break;
 	case ofi_op_atomic_fetch:
 		if (cmd->msg.hdr.atomic_op == FI_ATOMIC_READ)
 			cmd->msg.hdr.size = ofi_total_iov_len(resultv, result_count);
@@ -123,6 +123,59 @@ static void smr_format_inject_atomic(struct smr_cmd *cmd,
 	}
 }
 
+static ssize_t smr_do_atomic_inject(struct smr_ep *ep, struct smr_region *peer_smr,
+			int64_t id, int64_t peer_id, uint32_t op,
+			uint64_t op_flags, enum fi_hmem_iface iface,
+			uint64_t device, uint8_t datatype, uint8_t atomic_op,
+			const struct iovec *iov, size_t iov_count,
+			const struct iovec *resultv, size_t result_count,
+			const struct iovec *compv, size_t comp_count,
+			size_t total_len, void *context, uint16_t smr_flags)
+{
+	struct smr_cmd *cmd;
+	struct smr_inject_buf *tx_buf;
+	struct smr_tx_entry *pend;
+	struct smr_resp *resp;
+
+	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
+	tx_buf = smr_freestack_pop(smr_inject_pool(peer_smr));
+
+	smr_generic_format(cmd, peer_id, op, 0, 0, op_flags);
+	smr_generic_atomic_format(cmd, datatype, atomic_op);
+	smr_format_inject_atomic(cmd, iface, device, iov, iov_count, 
+				 resultv, result_count, compv, comp_count,
+				 peer_smr, tx_buf);
+
+	if (smr_flags & SMR_RMA_REQ || op_flags & FI_DELIVERY_COMPLETE) {
+		if (ofi_cirque_isfull(smr_resp_queue(ep->region))) {
+			smr_freestack_push(smr_inject_pool(peer_smr), tx_buf);
+			return -FI_EAGAIN;
+		}
+		resp = ofi_cirque_next(smr_resp_queue(ep->region));
+		pend = ofi_freestack_pop(ep->pend_fs);
+		smr_format_pend_resp(pend, cmd, context, iface, device, resultv,
+				     result_count, op_flags, id, resp);
+		cmd->msg.hdr.data = smr_get_offset(ep->region, resp);
+		ofi_cirque_commit(smr_resp_queue(ep->region));
+	}
+
+	cmd->msg.hdr.op_flags |= smr_flags;
+	ofi_cirque_commit(smr_cmd_queue(peer_smr));
+	peer_smr->cmd_cnt--;
+
+	return FI_SUCCESS;
+}
+
+static int smr_select_atomic_proto(uint32_t op, uint64_t total_len,
+				   uint64_t op_flags)
+{
+	if (op == ofi_op_atomic_compare || op == ofi_op_atomic_fetch ||
+	    op_flags & FI_DELIVERY_COMPLETE || total_len > SMR_MSG_DATA_LEN)
+		return smr_src_inject;
+
+	return smr_src_inline;
+}
+
 static ssize_t smr_generic_atomic(struct smr_ep *ep,
 			const struct fi_ioc *ioc, void **desc, size_t count,
 			const struct fi_ioc *compare_ioc, void **compare_desc,
@@ -133,18 +186,16 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 			enum fi_op atomic_op, void *context, uint32_t op,
 			uint64_t op_flags)
 {
-	struct smr_region *peer_smr;
-	struct smr_inject_buf *tx_buf;
-	struct smr_tx_entry *pend;
-	struct smr_resp *resp = NULL;
 	struct smr_cmd *cmd;
+	struct smr_region *peer_smr;
 	struct iovec iov[SMR_IOV_LIMIT];
 	struct iovec compare_iov[SMR_IOV_LIMIT];
 	struct iovec result_iov[SMR_IOV_LIMIT];
 	enum fi_hmem_iface iface;
 	uint64_t device;
-	int id, peer_id, err = 0;
-	uint16_t flags = 0;
+	uint16_t smr_flags = 0;
+	int64_t id, peer_id;
+	int proto;
 	ssize_t ret = 0;
 	size_t total_len;
 
@@ -153,29 +204,22 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 	assert(compare_count <= SMR_IOV_LIMIT);
 	assert(rma_count <= SMR_IOV_LIMIT);
 
-	id = (int) addr;
-	peer_id = smr_peer_data(ep->region)[id].addr.addr;
-
-	ret = smr_verify_peer(ep, id);
-	if (ret)
-		return ret;
+	id = smr_verify_peer(ep, addr);
+	if (id < 0)
+		return -FI_EAGAIN;
 
+	peer_id = smr_peer_data(ep->region)[id].addr.id;
 	peer_smr = smr_peer_region(ep->region, id);
-	fastlock_acquire(&peer_smr->lock);
+
+	pthread_spin_lock(&peer_smr->lock);
 	if (peer_smr->cmd_cnt < 2 || smr_peer_data(ep->region)[id].sar_status) {
 		ret = -FI_EAGAIN;
 		goto unlock_region;
 	}
 
-	fastlock_acquire(&ep->util_ep.tx_cq->cq_lock);
-	if (ofi_cirque_isfull(ep->util_ep.tx_cq->cirq)) {
-		ret = -FI_EAGAIN;
-		goto unlock_cq;
-	}
-
-	cmd = ofi_cirque_tail(smr_cmd_queue(peer_smr));
+	ofi_spin_lock(&ep->tx_lock);
 	total_len = ofi_datatype_size(datatype) * ofi_total_ioc_cnt(ioc, count);
-	
+
 	switch (op) {
 	case ofi_op_atomic_compare:
 		assert(compare_ioc);
@@ -187,7 +231,7 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 		assert(result_ioc);
 		ofi_ioc_to_iov(result_ioc, result_iov, result_count,
 			       ofi_datatype_size(datatype));
-		flags |= SMR_RMA_REQ;
+		smr_flags = SMR_RMA_REQ;
 		/* fall through */
 	case ofi_op_atomic:
 		if (atomic_op != FI_ATOMIC_READ) {
@@ -203,59 +247,39 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 
 	iface = smr_get_mr_hmem_iface(ep->util_ep.domain, desc, &device);
 
-	smr_generic_format(cmd, peer_id, op, 0, 0, op_flags);
-	smr_generic_atomic_format(cmd, datatype, atomic_op);
+	proto = smr_select_atomic_proto(op, total_len, op_flags);
 
-	if (total_len <= SMR_MSG_DATA_LEN && !(flags & SMR_RMA_REQ) &&
-	    !(op_flags & FI_DELIVERY_COMPLETE)) {
-		smr_format_inline_atomic(cmd, iface, device, iov, count, compare_iov,
-					 compare_count);
-	} else if (total_len <= SMR_INJECT_SIZE) {
-		tx_buf = smr_freestack_pop(smr_inject_pool(peer_smr));
-		smr_format_inject_atomic(cmd, iface, device, iov, count, result_iov,
-					 result_count, compare_iov, compare_count,
-					 peer_smr, tx_buf);
-		if (flags & SMR_RMA_REQ || op_flags & FI_DELIVERY_COMPLETE) {
-			if (ofi_cirque_isfull(smr_resp_queue(ep->region))) {
-				smr_freestack_push(smr_inject_pool(peer_smr), tx_buf);
-				ret = -FI_EAGAIN;
-				goto unlock_cq;
-			}
-			resp = ofi_cirque_tail(smr_resp_queue(ep->region));
-			pend = freestack_pop(ep->pend_fs);
-			smr_format_pend_resp(pend, cmd, context, iface, device, result_iov,
-					     result_count, id, resp);
-			cmd->msg.hdr.data = smr_get_offset(ep->region, resp);
-			ofi_cirque_commit(smr_resp_queue(ep->region));
-		}
+	if (proto == smr_src_inline) {
+		smr_do_atomic_inline(ep, peer_smr, id, peer_id, ofi_op_atomic,
+			 	     op_flags, iface, device, datatype, atomic_op,
+				     iov, count, total_len);
 	} else {
-		FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
-			"message too large\n");
-		ret = -FI_EINVAL;
-		goto unlock_cq;
+		ret = smr_do_atomic_inject(ep, peer_smr, id, peer_id, op,
+				op_flags, iface, device, datatype, atomic_op,
+				iov, count, result_iov, result_count,
+				compare_iov, compare_count, total_len, context,
+				smr_flags);
+		if (ret)
+			goto unlock_cq;
 	}
-	cmd->msg.hdr.op_flags |= flags;
 
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-
-	if (!resp) {
-		ret = smr_complete_tx(ep, context, op, cmd->msg.hdr.op_flags,
-				      err);
+	if (!(smr_flags & SMR_RMA_REQ) && !(op_flags & FI_DELIVERY_COMPLETE)) {
+		ret = smr_complete_tx(ep, context, op, op_flags);
 		if (ret) {
 			FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
 				"unable to process tx completion\n");
 		}
 	}
 
-	cmd = ofi_cirque_tail(smr_cmd_queue(peer_smr));
+	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
 	smr_format_rma_ioc(cmd, rma_ioc, rma_count);
 	ofi_cirque_commit(smr_cmd_queue(peer_smr));
 	peer_smr->cmd_cnt--;
+	smr_signal(peer_smr);
 unlock_cq:
-	fastlock_release(&ep->util_ep.tx_cq->cq_lock);
+	ofi_spin_unlock(&ep->tx_lock);
 unlock_region:
-	fastlock_release(&peer_smr->lock);
+	pthread_spin_unlock(&peer_smr->lock);
 	return ret;
 }
 
@@ -319,37 +343,33 @@ static ssize_t smr_atomic_inject(struct fid_ep *ep_fid, const void *buf,
 			size_t count, fi_addr_t dest_addr, uint64_t addr,
 			uint64_t key, enum fi_datatype datatype, enum fi_op op)
 {
+	struct smr_cmd *cmd;
 	struct smr_ep *ep;
 	struct smr_region *peer_smr;
-	struct smr_inject_buf *tx_buf;
-	struct smr_cmd *cmd;
 	struct iovec iov;
 	struct fi_rma_ioc rma_ioc;
-	int id, peer_id;
+	int64_t id, peer_id;
 	ssize_t ret = 0;
 	size_t total_len;
 
-	assert(count <= SMR_INJECT_SIZE);
-
 	ep = container_of(ep_fid, struct smr_ep, util_ep.ep_fid.fid);
 
-	id = (int) dest_addr;
-	peer_id = smr_peer_data(ep->region)[id].addr.addr;
-
-	ret = smr_verify_peer(ep, id);
-	if (ret)
-		return ret;
+	id = smr_verify_peer(ep, dest_addr);
+	if (id < 0)
+		return -FI_EAGAIN;
 
+	peer_id = smr_peer_data(ep->region)[id].addr.id;
 	peer_smr = smr_peer_region(ep->region, id);
-	fastlock_acquire(&peer_smr->lock);
+
+	pthread_spin_lock(&peer_smr->lock);
 	if (peer_smr->cmd_cnt < 2 || smr_peer_data(ep->region)[id].sar_status) {
 		ret = -FI_EAGAIN;
 		goto unlock_region;
 	}
 
-	cmd = ofi_cirque_tail(smr_cmd_queue(peer_smr));
 	total_len = count * ofi_datatype_size(datatype);
-	
+	assert(total_len <= SMR_INJECT_SIZE);
+
 	iov.iov_base = (void *) buf;
 	iov.iov_len = total_len;
 
@@ -357,27 +377,28 @@ static ssize_t smr_atomic_inject(struct fid_ep *ep_fid, const void *buf,
 	rma_ioc.count = count;
 	rma_ioc.key = key;
 
-	smr_generic_format(cmd, peer_id, ofi_op_atomic, 0, 0, 0);
-	smr_generic_atomic_format(cmd, datatype, op);
-
 	if (total_len <= SMR_MSG_DATA_LEN) {
-		smr_format_inline_atomic(cmd, FI_HMEM_SYSTEM, 0, &iov, 1, NULL, 0);
+		smr_do_atomic_inline(ep, peer_smr, id, peer_id, ofi_op_atomic,
+			 	     0, FI_HMEM_SYSTEM, 0, datatype, op,
+				     &iov, 1, total_len);
 	} else if (total_len <= SMR_INJECT_SIZE) {
-		tx_buf = smr_freestack_pop(smr_inject_pool(peer_smr));
-		smr_format_inject_atomic(cmd, FI_HMEM_SYSTEM, 0, &iov, 1, NULL,
-					 0, NULL, 0, peer_smr, tx_buf);
+		ret = smr_do_atomic_inject(ep, peer_smr, id, peer_id,
+				ofi_op_atomic, 0, FI_HMEM_SYSTEM, 0, datatype,
+				op, &iov, 1, NULL, 0, NULL, 0, total_len,
+				NULL, 0);
+		if (ret)
+			goto unlock_region;
 	}
 
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-	cmd = ofi_cirque_tail(smr_cmd_queue(peer_smr));
+	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
 	smr_format_rma_ioc(cmd, &rma_ioc, 1);
 	ofi_cirque_commit(smr_cmd_queue(peer_smr));
 	peer_smr->cmd_cnt--;
+	smr_signal(peer_smr);
 
 	ofi_ep_tx_cntr_inc_func(&ep->util_ep, ofi_op_atomic);
 unlock_region:
-	fastlock_release(&peer_smr->lock);
+	pthread_spin_unlock(&peer_smr->lock);
 	return ret;
 }
 
@@ -532,7 +553,13 @@ int smr_query_atomic(struct fid_domain *domain, enum fi_datatype datatype,
 	if (flags & FI_TAGGED) {
 		FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
 			"tagged atomic op not supported\n");
-		return -FI_EINVAL;
+		return -FI_EOPNOTSUPP;
+	}
+
+	if ((datatype == FI_INT128) || (datatype == FI_UINT128)) {
+		FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
+			"128-bit datatypes not supported\n");
+		return -FI_EOPNOTSUPP;
 	}
 
 	ret = ofi_atomic_valid(&smr_prov, datatype, op, flags);
