@@ -219,7 +219,7 @@ static void udpx_tx_comp(struct udpx_ep *ep, void *context)
 {
 	struct fi_cq_tagged_entry *comp;
 
-	comp = ofi_cirque_tail(ep->util_ep.tx_cq->cirq);
+	comp = ofi_cirque_next(ep->util_ep.tx_cq->cirq);
 	comp->op_context = context;
 	comp->flags = FI_SEND;
 	comp->len = 0;
@@ -239,7 +239,7 @@ static void udpx_rx_comp(struct udpx_ep *ep, void *context, uint64_t flags,
 {
 	struct fi_cq_tagged_entry *comp;
 
-	comp = ofi_cirque_tail(ep->util_ep.rx_cq->cirq);
+	comp = ofi_cirque_next(ep->util_ep.rx_cq->cirq);
 	comp->op_context = context;
 	comp->flags = FI_RECV | flags;
 	comp->len = len;
@@ -285,7 +285,7 @@ static void udpx_ep_progress(struct util_ep *util_ep)
 	hdr.msg_controllen = 0;
 	hdr.msg_flags = 0;
 
-	fastlock_acquire(&ep->util_ep.rx_cq->cq_lock);
+	ofi_genlock_lock(&ep->util_ep.rx_cq->cq_lock);
 	if (ofi_cirque_isempty(ep->rxq))
 		goto out;
 
@@ -299,7 +299,7 @@ static void udpx_ep_progress(struct util_ep *util_ep)
 		ofi_cirque_discard(ep->rxq);
 	}
 out:
-	fastlock_release(&ep->util_ep.rx_cq->cq_lock);
+	ofi_genlock_unlock(&ep->util_ep.rx_cq->cq_lock);
 }
 
 static ssize_t udpx_recvmsg(struct fid_ep *ep_fid, const struct fi_msg *msg,
@@ -310,13 +310,13 @@ static ssize_t udpx_recvmsg(struct fid_ep *ep_fid, const struct fi_msg *msg,
 	ssize_t ret;
 
 	ep = container_of(ep_fid, struct udpx_ep, util_ep.ep_fid.fid);
-	fastlock_acquire(&ep->util_ep.rx_cq->cq_lock);
+	ofi_genlock_lock(&ep->util_ep.rx_cq->cq_lock);
 	if (ofi_cirque_isfull(ep->rxq)) {
 		ret = -FI_EAGAIN;
 		goto out;
 	}
 
-	entry = ofi_cirque_tail(ep->rxq);
+	entry = ofi_cirque_next(ep->rxq);
 	entry->context = msg->context;
 	for (entry->iov_count = 0; entry->iov_count < msg->iov_count;
 	     entry->iov_count++) {
@@ -327,7 +327,7 @@ static ssize_t udpx_recvmsg(struct fid_ep *ep_fid, const struct fi_msg *msg,
 	ofi_cirque_commit(ep->rxq);
 	ret = 0;
 out:
-	fastlock_release(&ep->util_ep.rx_cq->cq_lock);
+	ofi_genlock_unlock(&ep->util_ep.rx_cq->cq_lock);
 	return ret;
 }
 
@@ -351,13 +351,13 @@ static ssize_t udpx_recv(struct fid_ep *ep_fid, void *buf, size_t len,
 	ssize_t ret;
 
 	ep = container_of(ep_fid, struct udpx_ep, util_ep.ep_fid.fid);
-	fastlock_acquire(&ep->util_ep.rx_cq->cq_lock);
+	ofi_genlock_lock(&ep->util_ep.rx_cq->cq_lock);
 	if (ofi_cirque_isfull(ep->rxq)) {
 		ret = -FI_EAGAIN;
 		goto out;
 	}
 
-	entry = ofi_cirque_tail(ep->rxq);
+	entry = ofi_cirque_next(ep->rxq);
 	entry->context = context;
 	entry->iov_count = 1;
 	entry->iov[0].iov_base = buf;
@@ -367,7 +367,7 @@ static ssize_t udpx_recv(struct fid_ep *ep_fid, void *buf, size_t len,
 	ofi_cirque_commit(ep->rxq);
 	ret = 0;
 out:
-	fastlock_release(&ep->util_ep.rx_cq->cq_lock);
+	ofi_genlock_unlock(&ep->util_ep.rx_cq->cq_lock);
 	return ret;
 }
 
@@ -392,7 +392,7 @@ static ssize_t udpx_sendto(struct udpx_ep *ep, const void *buf, size_t len,
 {
 	ssize_t ret;
 
-	fastlock_acquire(&ep->util_ep.tx_cq->cq_lock);
+	ofi_genlock_lock(&ep->util_ep.tx_cq->cq_lock);
 	if (ofi_cirque_isfull(ep->util_ep.tx_cq->cirq)) {
 		ret = -FI_EAGAIN;
 		goto out;
@@ -407,7 +407,7 @@ static ssize_t udpx_sendto(struct udpx_ep *ep, const void *buf, size_t len,
 		ret = -errno;
 	}
 out:
-	fastlock_release(&ep->util_ep.tx_cq->cq_lock);
+	ofi_genlock_unlock(&ep->util_ep.tx_cq->cq_lock);
 	return ret;
 }
 
@@ -448,7 +448,7 @@ static ssize_t udpx_sendmsg(struct fid_ep *ep_fid, const struct fi_msg *msg,
 	hdr.msg_controllen = 0;
 	hdr.msg_flags = 0;
 
-	fastlock_acquire(&ep->util_ep.tx_cq->cq_lock);
+	ofi_genlock_lock(&ep->util_ep.tx_cq->cq_lock);
 	if (ofi_cirque_isfull(ep->util_ep.tx_cq->cirq)) {
 		ret = -FI_EAGAIN;
 		goto out;
@@ -462,7 +462,7 @@ static ssize_t udpx_sendmsg(struct fid_ep *ep_fid, const struct fi_msg *msg,
 		ret = -errno;
 	}
 out:
-	fastlock_release(&ep->util_ep.tx_cq->cq_lock);
+	ofi_genlock_unlock(&ep->util_ep.tx_cq->cq_lock);
 	return ret;
 }
 
@@ -698,7 +698,8 @@ static int udpx_ep_ctrl(struct fid *fid, int command, void *arg)
 	ep = container_of(fid, struct udpx_ep, util_ep.ep_fid.fid);
 	switch (command) {
 	case FI_ENABLE:
-		if (!ep->util_ep.rx_cq || !ep->util_ep.tx_cq)
+		if ((ofi_needs_rx(ep->util_ep.caps) && !ep->util_ep.rx_cq) ||
+		    (ofi_needs_tx(ep->util_ep.caps) && !ep->util_ep.tx_cq))
 			return -FI_ENOCQ;
 		if (!ep->util_ep.av)
 			return -FI_ENOAV;
