@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2019, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -52,8 +52,8 @@
 #include <netdb.h>
 
 #include "rdma/fi_errno.h"
-#include "fi_enosys.h"
-#include "fi.h"
+#include "ofi_enosys.h"
+#include "ofi.h"
 
 #include "usnic_direct.h"
 #include "usnic_ip_utils.h"
@@ -65,7 +65,6 @@
 #include "usdf_av.h"
 #include "usdf_cm.h"
 #include "usdf_timer.h"
-#include "usdf_rdm.h"
 
 #include "fi_ext_usnic.h"
 
@@ -77,28 +76,14 @@ static int usdf_av_alloc_dest(struct usdf_dest **dest_o)
 	if (dest == NULL)
 		return -errno;
 
-	SLIST_INIT(&dest->ds_rdm_rdc_list);
-
 	*dest_o = dest;
 	return 0;
 }
 
 static void usdf_av_free_dest(struct usdf_dest *dest)
 {
-	struct usdf_rdm_connection *rdc = NULL;
-
 	LIST_REMOVE(dest, ds_addresses_entry);
 
-	while (!SLIST_EMPTY(&dest->ds_rdm_rdc_list)) {
-		rdc = SLIST_FIRST(&dest->ds_rdm_rdc_list);
-		rdc->dc_dest = NULL;
-
-		SLIST_REMOVE(&dest->ds_rdm_rdc_list, rdc, usdf_rdm_connection,
-			     dc_addr_link);
-		if (rdc)
-			rdc->dc_dest = NULL;
-	}
-
 	free(dest);
 }
 
@@ -363,6 +348,10 @@ usdf_am_insert_async(struct fid_av *fav, const void *addr, size_t count,
 
 		if (addr_format_str) {
 			usdf_str_toaddr(addr_str[i], &cur_sin);
+			if (NULL == cur_sin) {
+				ret = -FI_ENOMEM;
+				goto fail;
+			}
 			sin = cur_sin;
 		}
 
@@ -469,6 +458,12 @@ usdf_am_insert_sync(struct fid_av *fav, const void *addr, size_t count,
 
 		if (addr_format_str) {
 			usdf_str_toaddr(addr_str[i], &cur_sin);
+			if (NULL == cur_sin) {
+				if (flags & FI_SYNC_ERR)
+					errors[i] = -ENOMEM;
+
+				return ret_count;
+			}
 			sin = cur_sin;
 		}
 
@@ -824,7 +819,7 @@ fi_addr_t usdf_av_lookup_addr(struct usdf_av *av,
 /* Return sockaddr_in pointer. Must be used with usdf_free_sin_if_needed()
  * to cleanup properly.
  */
-struct sockaddr_in *usdf_format_to_sin(struct fi_info *info, const void *addr)
+struct sockaddr_in *usdf_format_to_sin(const struct fi_info *info, const void *addr)
 {
 	struct sockaddr_in *sin;
 
@@ -846,7 +841,7 @@ struct sockaddr_in *usdf_format_to_sin(struct fi_info *info, const void *addr)
 
 /* Utility function to free the sockaddr_in allocated from usdf_format_to_sin()
  */
-void usdf_free_sin_if_needed(struct fi_info *info, struct sockaddr_in *sin)
+void usdf_free_sin_if_needed(const struct fi_info *info, struct sockaddr_in *sin)
 {
 	if (info && info->addr_format == FI_ADDR_STR)
 		free(sin);
@@ -855,7 +850,7 @@ void usdf_free_sin_if_needed(struct fi_info *info, struct sockaddr_in *sin)
 /* Convert sockaddr_in pointer to appropriate format.
  * If conversion happens, destroy the origin. (to minimize cleaning up code)
  */
-void *usdf_sin_to_format(struct fi_info *info, void *addr, size_t *len)
+void *usdf_sin_to_format(const struct fi_info *info, void *addr, size_t *len)
 {
 	size_t addr_strlen;
 	char *addrstr;
