@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2016-2017 Intel Corporation. All rights reserved.
+ * Copyright (c) 2016-2021 Intel Corporation. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -39,6 +39,7 @@
 #include <sys/mman.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <stdio.h>
 
 #include <ofi_shm.h>
 
@@ -61,51 +62,64 @@ void smr_cleanup(void)
 static void smr_peer_addr_init(struct smr_addr *peer)
 {
 	memset(peer->name, 0, SMR_NAME_MAX);
-	peer->addr = FI_ADDR_UNSPEC;
+	peer->id = -1;
 }
 
 void smr_cma_check(struct smr_region *smr, struct smr_region *peer_smr)
 {
 	struct iovec local_iov, remote_iov;
+	int remote_pid;
 	int ret;
 
-	if (peer_smr->cma_cap != SMR_CMA_CAP_NA) {
-		smr->cma_cap = peer_smr->cma_cap;
+	if (smr != peer_smr && peer_smr->cma_cap_peer != SMR_CMA_CAP_NA) {
+		smr->cma_cap_peer = peer_smr->cma_cap_peer;
 		return;
 	}
-	local_iov.iov_base = &smr->cma_cap;
-	local_iov.iov_len = sizeof(smr->cma_cap);
+	remote_pid = peer_smr->pid;
+	local_iov.iov_base = &remote_pid;
+	local_iov.iov_len = sizeof(remote_pid);
 	remote_iov.iov_base = (char *)peer_smr->base_addr +
-			      ((char *)&peer_smr->cma_cap - (char *)peer_smr);
-	remote_iov.iov_len = sizeof(peer_smr->cma_cap);
+			      ((char *)&peer_smr->pid - (char *)peer_smr);
+	remote_iov.iov_len = sizeof(peer_smr->pid);
 	ret = ofi_process_vm_writev(peer_smr->pid, &local_iov, 1,
 				    &remote_iov, 1, 0);
-	smr->cma_cap = (ret == -1) ? SMR_CMA_CAP_OFF : SMR_CMA_CAP_ON;
-	peer_smr->cma_cap = smr->cma_cap;
+	assert(remote_pid == peer_smr->pid);
+
+	if (smr == peer_smr) {
+		smr->cma_cap_self = (ret == -1) ? SMR_CMA_CAP_OFF : SMR_CMA_CAP_ON;
+	} else {
+		smr->cma_cap_peer = (ret == -1) ? SMR_CMA_CAP_OFF : SMR_CMA_CAP_ON;
+		peer_smr->cma_cap_peer = smr->cma_cap_peer;
+	}
 }
 
 size_t smr_calculate_size_offsets(size_t tx_count, size_t rx_count,
 				  size_t *cmd_offset, size_t *resp_offset,
 				  size_t *inject_offset, size_t *sar_offset,
-				  size_t *peer_offset, size_t *name_offset)
+				  size_t *peer_offset, size_t *name_offset,
+				  size_t *sock_offset)
 {
 	size_t cmd_queue_offset, resp_queue_offset, inject_pool_offset;
 	size_t sar_pool_offset, peer_data_offset, ep_name_offset;
-	size_t tx_size, rx_size, total_size;
+	size_t tx_size, rx_size, total_size, sock_name_offset;
 
 	tx_size = roundup_power_of_two(tx_count);
 	rx_size = roundup_power_of_two(rx_count);
 
-	cmd_queue_offset = sizeof(struct smr_region);
+	/* Align cmd_queue offset to 128-bit boundary. */
+	cmd_queue_offset = ofi_get_aligned_size(sizeof(struct smr_region), 16);
 	resp_queue_offset = cmd_queue_offset + sizeof(struct smr_cmd_queue) +
 			    sizeof(struct smr_cmd) * rx_size;
 	inject_pool_offset = resp_queue_offset + sizeof(struct smr_resp_queue) +
 			     sizeof(struct smr_resp) * tx_size;
-	sar_pool_offset = inject_pool_offset + sizeof(struct smr_inject_pool) +
-			  sizeof(struct smr_inject_pool_entry) * rx_size;
-	peer_data_offset = sar_pool_offset + sizeof(struct smr_sar_pool) +
-			   sizeof(struct smr_sar_pool_entry) * SMR_MAX_PEERS;
-	ep_name_offset = peer_data_offset + sizeof(struct smr_peer_data) * SMR_MAX_PEERS;
+	sar_pool_offset = inject_pool_offset +
+		freestack_size(sizeof(struct smr_inject_buf), rx_size);
+	peer_data_offset = sar_pool_offset +
+		freestack_size(sizeof(struct smr_sar_buf), SMR_MAX_PEERS);
+	ep_name_offset = peer_data_offset + sizeof(struct smr_peer_data) *
+		SMR_MAX_PEERS;
+
+	sock_name_offset = ep_name_offset + SMR_NAME_MAX;
 
 	if (cmd_offset)
 		*cmd_offset = cmd_queue_offset;
@@ -119,8 +133,10 @@ size_t smr_calculate_size_offsets(size_t tx_count, size_t rx_count,
 		*peer_offset = peer_data_offset;
 	if (name_offset)
 		*name_offset = ep_name_offset;
+	if (sock_offset)
+		*sock_offset = sock_name_offset;
 
-	total_size = ep_name_offset + SMR_NAME_MAX;
+	total_size = sock_name_offset + SMR_SOCK_NAME_MAX;
 
 	/*
  	 * Revisit later to see if we really need the size adjustment, or
@@ -131,6 +147,50 @@ size_t smr_calculate_size_offsets(size_t tx_count, size_t rx_count,
 	return total_size;
 }
 
+static int smr_retry_map(const char *name, int *fd)
+{
+	char tmp[NAME_MAX];
+	struct smr_region *old_shm;
+	struct stat sts;
+	int shm_pid;
+
+	*fd = shm_open(name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
+	if (*fd < 0)
+		return -errno;
+
+	old_shm = mmap(NULL, sizeof(*old_shm), PROT_READ | PROT_WRITE,
+		       MAP_SHARED, *fd, 0);
+	if (old_shm == MAP_FAILED)
+		goto err;
+
+        /* No backwards compatibility for now. */
+	if (old_shm->version != SMR_VERSION) {
+		munmap(old_shm, sizeof(*old_shm));
+		goto err;
+	}
+	shm_pid = old_shm->pid;
+	munmap(old_shm, sizeof(*old_shm));
+
+	if (!shm_pid)
+		return FI_SUCCESS;
+
+	memset(tmp, 0, sizeof(tmp));
+	snprintf(tmp, sizeof(tmp), "/proc/%d", shm_pid);
+
+	if (stat(tmp, &sts) == -1 && errno == ENOENT)
+		return FI_SUCCESS;
+
+err:
+	close(*fd);
+	shm_unlink(name);
+	return -FI_EBUSY;
+}
+
+static void smr_lock_init(pthread_spinlock_t *lock)
+{
+	pthread_spin_init(lock, PTHREAD_PROCESS_SHARED);
+}
+
 /* TODO: Determine if aligning SMR data helps performance */
 int smr_create(const struct fi_provider *prov, struct smr_map *map,
 	       const struct smr_attr *attr, struct smr_region *volatile *smr)
@@ -138,7 +198,7 @@ int smr_create(const struct fi_provider *prov, struct smr_map *map,
 	struct smr_ep_name *ep_name;
 	size_t total_size, cmd_queue_offset, peer_data_offset;
 	size_t resp_queue_offset, inject_pool_offset, name_offset;
-	size_t sar_pool_offset;
+	size_t sar_pool_offset, sock_name_offset;
 	int fd, ret, i;
 	void *mapped_addr;
 	size_t tx_size, rx_size;
@@ -148,12 +208,25 @@ int smr_create(const struct fi_provider *prov, struct smr_map *map,
 	total_size = smr_calculate_size_offsets(tx_size, rx_size, &cmd_queue_offset,
 					&resp_queue_offset, &inject_pool_offset,
 					&sar_pool_offset, &peer_data_offset,
-					&name_offset);
+					&name_offset, &sock_name_offset);
 
-	fd = shm_open(attr->name, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);
+	fd = shm_open(attr->name, O_RDWR | O_CREAT | O_EXCL, S_IRUSR | S_IWUSR);
 	if (fd < 0) {
-		FI_WARN(prov, FI_LOG_EP_CTRL, "shm_open error\n");
-		return -errno;
+		if (errno != EEXIST) {
+			FI_WARN(prov, FI_LOG_EP_CTRL,
+				"shm_open error (%s): %s\n",
+				attr->name, strerror(errno));
+			return -errno;
+		}
+
+		ret = smr_retry_map(attr->name, &fd);
+		if (ret) {
+			FI_WARN(prov, FI_LOG_EP_CTRL, "shm file in use (%s)\n",
+				attr->name);
+			return ret;
+		}
+		FI_WARN(prov, FI_LOG_EP_CTRL,
+			"Overwriting shm from dead process (%s)\n", attr->name);
 	}
 
 	ep_name = calloc(1, sizeof(*ep_name));
@@ -189,14 +262,14 @@ int smr_create(const struct fi_provider *prov, struct smr_map *map,
 	pthread_mutex_unlock(&ep_list_lock);
 
 	*smr = mapped_addr;
-	fastlock_init(&(*smr)->lock);
-	fastlock_acquire(&(*smr)->lock);
+	smr_lock_init(&(*smr)->lock);
+	ofi_atomic_initialize32(&(*smr)->signal, 0);
 
 	(*smr)->map = map;
 	(*smr)->version = SMR_VERSION;
 	(*smr)->flags = SMR_FLAG_ATOMIC | SMR_FLAG_DEBUG;
-	(*smr)->pid = getpid();
-	(*smr)->cma_cap = SMR_CMA_CAP_NA;
+	(*smr)->cma_cap_peer = SMR_CMA_CAP_NA;
+	(*smr)->cma_cap_self = SMR_CMA_CAP_NA;
 	(*smr)->base_addr = *smr;
 
 	(*smr)->total_size = total_size;
@@ -206,22 +279,28 @@ int smr_create(const struct fi_provider *prov, struct smr_map *map,
 	(*smr)->sar_pool_offset = sar_pool_offset;
 	(*smr)->peer_data_offset = peer_data_offset;
 	(*smr)->name_offset = name_offset;
+	(*smr)->sock_name_offset = sock_name_offset;
 	(*smr)->cmd_cnt = rx_size;
 	/* Limit of 1 outstanding SAR message per peer */
 	(*smr)->sar_cnt = SMR_MAX_PEERS;
+	(*smr)->max_sar_buf_per_peer = SMR_BUF_BATCH_MAX;
 
 	smr_cmd_queue_init(smr_cmd_queue(*smr), rx_size);
 	smr_resp_queue_init(smr_resp_queue(*smr), tx_size);
-	smr_inject_pool_init(smr_inject_pool(*smr), rx_size);
-	smr_sar_pool_init(smr_sar_pool(*smr), SMR_MAX_PEERS);
+	smr_freestack_init(smr_inject_pool(*smr), rx_size,
+			sizeof(struct smr_inject_buf));
+	smr_freestack_init(smr_sar_pool(*smr), SMR_MAX_PEERS,
+			sizeof(struct smr_sar_buf));
 	for (i = 0; i < SMR_MAX_PEERS; i++) {
 		smr_peer_addr_init(&smr_peer_data(*smr)[i].addr);
 		smr_peer_data(*smr)[i].sar_status = 0;
+		smr_peer_data(*smr)[i].name_sent = 0;
 	}
 
 	strncpy((char *) smr_name(*smr), attr->name, total_size - name_offset);
-	fastlock_release(&(*smr)->lock);
 
+	/* Must be set last to signal full initialization to peers */
+	(*smr)->pid = getpid();
 	return 0;
 
 remove:
@@ -240,6 +319,16 @@ void smr_free(struct smr_region *smr)
 	munmap(smr, smr->total_size);
 }
 
+static int smr_name_compare(struct ofi_rbmap *map, void *key, void *data)
+{
+	struct smr_map *smr_map;
+
+	smr_map = container_of(map, struct smr_map, rbmap);
+
+	return strncmp(smr_map->peers[(uintptr_t) data].peer.name,
+		       (char *) key, SMR_NAME_MAX);
+}
+
 int smr_map_create(const struct fi_provider *prov, int peer_count,
 		   struct smr_map **map)
 {
@@ -251,10 +340,13 @@ int smr_map_create(const struct fi_provider *prov, int peer_count,
 		return -FI_ENOMEM;
 	}
 
-	for (i = 0; i < peer_count; i++)
+	for (i = 0; i < peer_count; i++) {
 		smr_peer_addr_init(&(*map)->peers[i].peer);
+		(*map)->peers[i].fiaddr = FI_ADDR_UNSPEC;
+	}
 
-	fastlock_init(&(*map)->lock);
+	ofi_rbmap_init(&(*map)->rbmap, smr_name_compare);
+	ofi_spin_init(&(*map)->lock);
 
 	return 0;
 }
@@ -270,11 +362,16 @@ int smr_map_to_region(const struct fi_provider *prov, struct smr_peer *peer_buf)
 	struct smr_region *peer;
 	size_t size;
 	int fd, ret = 0;
+	struct stat sts;
 	struct dlist_entry *entry;
+	const char *name = smr_no_prefix(peer_buf->peer.name);
+	char tmp[SMR_PATH_MAX];
+
+	if (peer_buf->region)
+		return FI_SUCCESS;
 
 	pthread_mutex_lock(&ep_list_lock);
-	entry = dlist_find_first_match(&ep_name_list, smr_match_name,
-				       peer_buf->peer.name);
+	entry = dlist_find_first_match(&ep_name_list, smr_match_name, name);
 	if (entry) {
 		peer_buf->region = container_of(entry, struct smr_ep_name,
 						entry)->region;
@@ -283,12 +380,24 @@ int smr_map_to_region(const struct fi_provider *prov, struct smr_peer *peer_buf)
 	}
 	pthread_mutex_unlock(&ep_list_lock);
 
-	fd = shm_open(peer_buf->peer.name, O_RDWR, S_IRUSR | S_IWUSR);
+	fd = shm_open(name, O_RDWR, S_IRUSR | S_IWUSR);
 	if (fd < 0) {
 		FI_WARN_ONCE(prov, FI_LOG_AV, "shm_open error\n");
 		return -errno;
 	}
 
+	memset(tmp, 0, sizeof(tmp));
+	snprintf(tmp, sizeof(tmp), "%s%s", SMR_DIR, name);
+	if (stat(tmp, &sts) == -1) {
+		ret = -errno;
+		goto out;
+	}
+
+	if (sts.st_size < sizeof(*peer)) {
+		ret = -FI_ENOENT;
+		goto out;
+	}
+
 	peer = mmap(NULL, sizeof(*peer), PROT_READ | PROT_WRITE,
 		    MAP_SHARED, fd, 0);
 	if (peer == MAP_FAILED) {
@@ -300,7 +409,7 @@ int smr_map_to_region(const struct fi_provider *prov, struct smr_peer *peer_buf)
 	if (!peer->pid) {
 		FI_WARN(prov, FI_LOG_AV, "peer not initialized\n");
 		munmap(peer, sizeof(*peer));
-		ret = -FI_EAGAIN;
+		ret = -FI_ENOENT;
 		goto out;
 	}
 
@@ -315,108 +424,131 @@ out:
 	return ret;
 }
 
-void smr_map_to_endpoint(struct smr_region *region, int index)
+void smr_map_to_endpoint(struct smr_region *region, int64_t id)
 {
 	struct smr_region *peer_smr;
-	struct smr_peer_data *local_peers, *peer_peers;
-	int peer_index;
+	struct smr_peer_data *local_peers;
+
+	if (region->map->peers[id].peer.id < 0)
+		return;
 
 	local_peers = smr_peer_data(region);
 
-	strncpy(smr_peer_data(region)[index].addr.name,
-		region->map->peers[index].peer.name, SMR_NAME_MAX - 1);
-	smr_peer_data(region)[index].addr.name[SMR_NAME_MAX - 1] = '\0';
-	if (region->map->peers[index].peer.addr == FI_ADDR_UNSPEC)
-		return;
+	strncpy(local_peers[id].addr.name,
+		region->map->peers[id].peer.name, SMR_NAME_MAX - 1);
+	local_peers[id].addr.name[SMR_NAME_MAX - 1] = '\0';
 
-	peer_smr = smr_peer_region(region, index);
-	peer_peers = smr_peer_data(peer_smr);
+	peer_smr = smr_peer_region(region, id);
 
-	if (region->cma_cap == SMR_CMA_CAP_NA && region != peer_smr)
+	if ((region != peer_smr && region->cma_cap_peer == SMR_CMA_CAP_NA) ||
+	    (region == peer_smr && region->cma_cap_self == SMR_CMA_CAP_NA))
 		smr_cma_check(region, peer_smr);
-
-	for (peer_index = 0; peer_index < SMR_MAX_PEERS; peer_index++) {
-		if (!strncmp(smr_name(region),
-		    peer_peers[peer_index].addr.name, SMR_NAME_MAX))
-			break;
-	}
-	if (peer_index != SMR_MAX_PEERS) {
-		peer_peers[peer_index].addr.addr = index;
-		local_peers[index].addr.addr = peer_index;
-	}
 }
 
-void smr_unmap_from_endpoint(struct smr_region *region, int index)
+void smr_unmap_from_endpoint(struct smr_region *region, int64_t id)
 {
 	struct smr_region *peer_smr;
 	struct smr_peer_data *local_peers, *peer_peers;
-	int peer_index;
+	int64_t peer_id;
 
 	local_peers = smr_peer_data(region);
 
-	memset(local_peers[index].addr.name, 0, SMR_NAME_MAX);
-	peer_index = region->map->peers[index].peer.addr;
-	if (peer_index == FI_ADDR_UNSPEC)
+	memset(local_peers[id].addr.name, 0, SMR_NAME_MAX);
+	peer_id = region->map->peers[id].peer.id;
+	if (peer_id < 0)
 		return;
 
-	peer_smr = smr_peer_region(region, index);
+	peer_smr = smr_peer_region(region, id);
 	peer_peers = smr_peer_data(peer_smr);
 
-	peer_peers[peer_index].addr.addr = FI_ADDR_UNSPEC;
+	peer_peers[peer_id].addr.id = -1;
+	peer_peers[peer_id].name_sent = 0;
 }
 
 void smr_exchange_all_peers(struct smr_region *region)
 {
-	int i;
+	int64_t i;
 	for (i = 0; i < SMR_MAX_PEERS; i++)
 		smr_map_to_endpoint(region, i);
 }
 
 int smr_map_add(const struct fi_provider *prov, struct smr_map *map,
-		const char *name, int id)
+		const char *name, int64_t *id)
 {
-	int ret = 0;
+	struct ofi_rbnode *node;
+	int tries = 0, ret = 0;
 
-	fastlock_acquire(&map->lock);
-	strncpy(map->peers[id].peer.name, name, SMR_NAME_MAX);
-	map->peers[id].peer.name[SMR_NAME_MAX - 1] = '\0';
-	ret = smr_map_to_region(prov, &map->peers[id]);
+	ofi_spin_lock(&map->lock);
+	ret = ofi_rbmap_insert(&map->rbmap, (void *) name,
+			       (void *) (intptr_t) *id, &node);
+	if (ret) {
+		assert(ret == -FI_EALREADY);
+		*id = (intptr_t) node->data;
+		ofi_spin_unlock(&map->lock);
+		return 0;
+	}
+
+	while (map->peers[map->cur_id].peer.id != -1 &&
+	       tries < SMR_MAX_PEERS) {
+		if (++map->cur_id == SMR_MAX_PEERS)
+			map->cur_id = 0;
+		tries++;
+	}
+
+	assert(map->cur_id < SMR_MAX_PEERS && tries < SMR_MAX_PEERS);
+	*id = map->cur_id;
+	node->data = (void *) (intptr_t) *id;
+	strncpy(map->peers[*id].peer.name, name, SMR_NAME_MAX);
+	map->peers[*id].peer.name[SMR_NAME_MAX - 1] = '\0';
+	map->peers[*id].region = NULL;
+
+	ret = smr_map_to_region(prov, &map->peers[*id]);
 	if (!ret)
-		map->peers[id].peer.addr = id;
-	fastlock_release(&map->lock);
+		map->peers[*id].peer.id = *id;
 
+	map->num_peers++;
+	ofi_spin_unlock(&map->lock);
 	return ret == -ENOENT ? 0 : ret;
 }
 
-void smr_map_del(struct smr_map *map, int id)
+void smr_map_del(struct smr_map *map, int64_t id)
 {
 	struct dlist_entry *entry;
 
-	if (id >= SMR_MAX_PEERS || id < 0 ||
-	    map->peers[id].peer.addr == FI_ADDR_UNSPEC)
+	if (id >= SMR_MAX_PEERS || id < 0 || map->peers[id].peer.id < 0)
 		return;
 
 	pthread_mutex_lock(&ep_list_lock);
 	entry = dlist_find_first_match(&ep_name_list, smr_match_name,
-				       map->peers[id].peer.name);
+				       smr_no_prefix(map->peers[id].peer.name));
 	pthread_mutex_unlock(&ep_list_lock);
+
+	ofi_spin_lock(&map->lock);
 	if (!entry)
 		munmap(map->peers[id].region, map->peers[id].region->total_size);
 
-	map->peers[id].peer.addr = FI_ADDR_UNSPEC;
+	(void) ofi_rbmap_find_delete(&map->rbmap,
+				     (void *) map->peers[id].peer.name);
+
+	map->peers[id].fiaddr = FI_ADDR_UNSPEC;
+	map->peers[id].peer.id = -1;
+	map->num_peers--;
+
+	ofi_spin_unlock(&map->lock);
 }
 
 void smr_map_free(struct smr_map *map)
 {
-	int i;
+	int64_t i;
 
 	for (i = 0; i < SMR_MAX_PEERS; i++)
 		smr_map_del(map, i);
 
+	ofi_rbmap_cleanup(&map->rbmap);
 	free(map);
 }
 
-struct smr_region *smr_map_get(struct smr_map *map, int id)
+struct smr_region *smr_map_get(struct smr_map *map, int64_t id)
 {
 	if (id < 0 || id >= SMR_MAX_PEERS)
 		return NULL;
