@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2015 Intel Corporation, Inc.  All rights reserved.
+ * Copyright (c) 2013-2021 Intel Corporation, Inc.  All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -36,6 +36,7 @@
 
 #include "fi_verbs.h"
 #include "ofi_hmem.h"
+#include "verbs_osd.h"
 
 static void vrb_fini(void);
 
@@ -104,13 +105,15 @@ int vrb_sockaddr_len(struct sockaddr *addr)
 
 static int
 vrb_get_rdmacm_rai(const char *node, const char *service, uint64_t flags,
-		 const struct fi_info *hints, struct rdma_addrinfo **rai)
+		uint32_t addr_format, void *src_addr, size_t src_addrlen,
+		void *dest_addr, size_t dest_addrlen, struct rdma_addrinfo **rai)
 {
 	struct rdma_addrinfo rai_hints, *_rai;
 	struct rdma_addrinfo **cur, *next;
 	int ret;
 
-	ret = vrb_fi_to_rai(hints, flags, &rai_hints);
+	ret = vrb_set_rai(addr_format, src_addr, src_addrlen, dest_addr,
+				       dest_addrlen, flags, &rai_hints);
 	if (ret)
 		goto out;
 
@@ -120,9 +123,9 @@ vrb_get_rdmacm_rai(const char *node, const char *service, uint64_t flags,
 		rai_hints.ai_flags |= RAI_PASSIVE;
 	}
 
-	ret = rdma_getaddrinfo(node, service, &rai_hints, &_rai);
+	ret = rdma_getaddrinfo((char *) node, (char *) service, &rai_hints, &_rai);
 	if (ret) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_getaddrinfo", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_getaddrinfo");
 		if (errno)
 			ret = -errno;
 		goto out;
@@ -132,7 +135,7 @@ vrb_get_rdmacm_rai(const char *node, const char *service, uint64_t flags,
 	 * Remove ib_rai entries added by IBACM to
 	 * prevent wrong ib_connect_hdr from being sent in connect request.
 	 */
-	if (hints && (hints->addr_format != FI_SOCKADDR_IB)) {
+	if (addr_format && (addr_format != FI_SOCKADDR_IB)) {
 		for (cur = &_rai; *cur; ) {
 			if ((*cur)->ai_family == AF_IB) {
 				next = (*cur)->ai_next;
@@ -154,8 +157,10 @@ out:
 	return ret;
 }
 
-static int vrb_get_sib_rai(const char *node, const char *service, uint64_t flags,
-		      const struct fi_info *hints, struct rdma_addrinfo **rai)
+static int
+vrb_get_sib_rai(const char *node, const char *service, uint64_t flags,
+		uint32_t addr_format, void *src_addr, size_t src_addrlen,
+		void *dest_addr, size_t dest_addrlen, struct rdma_addrinfo **rai)
 {
 	struct sockaddr_ib *sib;
 	size_t sib_len;
@@ -169,7 +174,8 @@ static int vrb_get_sib_rai(const char *node, const char *service, uint64_t flags
 	if (*rai == NULL)
 		return -FI_ENOMEM;
 
-	ret = vrb_fi_to_rai(hints, flags, *rai);
+	ret = vrb_set_rai(addr_format, src_addr, src_addrlen, dest_addr,
+			  dest_addrlen, flags, *rai);
 	if (ret)
 		return ret;
 
@@ -184,7 +190,7 @@ static int vrb_get_sib_rai(const char *node, const char *service, uint64_t flags
 
 		if (service) {
 			ret = asprintf(&straddr, "%s%s:%s", has_prefix ? "" : prefix,
-			                                    node, service);
+				       node, service);
 		} else {
 			ret = asprintf(&straddr, "%s%s", has_prefix ? "" : prefix, node);
 		}
@@ -234,15 +240,18 @@ static int vrb_get_sib_rai(const char *node, const char *service, uint64_t flags
 	return 0;
 }
 
-int vrb_get_rdma_rai(const char *node, const char *service, uint64_t flags,
-		 const struct fi_info *hints, struct rdma_addrinfo **rai)
+static int
+vrb_get_rdma_rai(const char *node, const char *service, uint32_t addr_format,
+		void *src_addr, size_t src_addrlen, void *dest_addr,
+		size_t dest_addrlen, uint64_t flags, struct rdma_addrinfo **rai)
 {
-	if (hints && hints->addr_format == FI_SOCKADDR_IB &&
-	    (node || hints->src_addr || hints->dest_addr)) {
-		return vrb_get_sib_rai(node, service, flags, hints, rai);
+	if (addr_format == FI_SOCKADDR_IB && (node || src_addr || dest_addr)) {
+		return vrb_get_sib_rai(node, service, flags, addr_format, src_addr,
+					src_addrlen, dest_addr, dest_addrlen, rai);
 	}
 
-	return vrb_get_rdmacm_rai(node, service, flags, hints, rai);
+	return vrb_get_rdmacm_rai(node, service, flags, addr_format, src_addr,
+				  src_addrlen, dest_addr, dest_addrlen, rai);
 }
 
 int vrb_get_rai_id(const char *node, const char *service, uint64_t flags,
@@ -252,13 +261,21 @@ int vrb_get_rai_id(const char *node, const char *service, uint64_t flags,
 	int ret;
 
 	// TODO create a similar function that won't require pruning ib_rai
-	ret = vrb_get_rdma_rai(node, service, flags, hints, rai);
+	if (hints) {
+		ret = vrb_get_rdma_rai(node, service, hints->addr_format, hints->src_addr,
+				       hints->src_addrlen, hints->dest_addr,
+				       hints->dest_addrlen, flags, rai);
+	} else {
+		ret = vrb_get_rdma_rai(node, service, FI_FORMAT_UNSPEC, NULL, 0, NULL,
+				       0, flags, rai);
+	}
 	if (ret)
 		return ret;
 
-	ret = rdma_create_id(NULL, id, NULL, vrb_get_port_space(hints));
+	ret = rdma_create_id(NULL, id, NULL, vrb_get_port_space(hints ? hints->addr_format:
+					FI_FORMAT_UNSPEC));
 	if (ret) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_create_id", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_create_id");
 		ret = -errno;
 		goto err1;
 	}
@@ -266,7 +283,7 @@ int vrb_get_rai_id(const char *node, const char *service, uint64_t flags,
 	if ((*rai)->ai_flags & RAI_PASSIVE) {
 		ret = rdma_bind_addr(*id, (*rai)->ai_src_addr);
 		if (ret) {
-			VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_bind_addr", errno);
+			VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_bind_addr");
 			ofi_straddr_log(&vrb_prov, FI_LOG_INFO, FI_LOG_FABRIC,
 					"bind addr", (*rai)->ai_src_addr);
 			ret = -errno;
@@ -278,7 +295,7 @@ int vrb_get_rai_id(const char *node, const char *service, uint64_t flags,
 	ret = rdma_resolve_addr(*id, (*rai)->ai_src_addr,
 				(*rai)->ai_dst_addr, VERBS_RESOLVE_TIMEOUT);
 	if (ret) {
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_resolve_addr", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_resolve_addr");
 		ofi_straddr_log(&vrb_prov, FI_LOG_INFO, FI_LOG_FABRIC,
 				"src addr", (*rai)->ai_src_addr);
 		ofi_straddr_log(&vrb_prov, FI_LOG_INFO, FI_LOG_FABRIC,
@@ -289,27 +306,29 @@ int vrb_get_rai_id(const char *node, const char *service, uint64_t flags,
 	return 0;
 err2:
 	if (rdma_destroy_id(*id))
-		VERBS_INFO_ERRNO(FI_LOG_FABRIC, "rdma_destroy_id", errno);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_destroy_id");
 err1:
 	rdma_freeaddrinfo(*rai);
 	return ret;
 }
 
-int vrb_create_ep(const struct fi_info *hints, enum rdma_port_space ps,
+int vrb_create_ep(struct vrb_ep *ep, enum rdma_port_space ps,
 		     struct rdma_cm_id **id)
 {
 	struct rdma_addrinfo *rai = NULL;
 	int ret;
 
-	ret = vrb_get_rdma_rai(NULL, NULL, 0, hints, &rai);
+	ret = vrb_get_rdma_rai(NULL, NULL, ep->info_attr.addr_format,
+				ep->info_attr.src_addr, ep->info_attr.src_addrlen,
+				ep->info_attr.dest_addr, ep->info_attr.dest_addrlen,
+				0, &rai);
 	if (ret) {
 		return ret;
 	}
 
 	if (rdma_create_id(NULL, id, NULL, ps)) {
 		ret = -errno;
-		FI_WARN(&vrb_prov, FI_LOG_FABRIC, "rdma_create_id failed: "
-			"%s (%d)\n", strerror(-ret), -ret);
+		VRB_WARN_ERRNO(FI_LOG_FABRIC, "rdma_create_id");
 		goto err1;
 	}
 
@@ -324,8 +343,7 @@ int vrb_create_ep(const struct fi_info *hints, enum rdma_port_space ps,
 	if (rdma_resolve_addr(*id, rai->ai_src_addr, rai->ai_dst_addr,
 			      VERBS_RESOLVE_TIMEOUT)) {
 		ret = -errno;
-		FI_WARN(&vrb_prov, FI_LOG_EP_CTRL, "rdma_resolve_addr failed: "
-			"%s (%d)\n", strerror(-ret), -ret);
+		VRB_WARN_ERRNO(FI_LOG_EP_CTRL, "rdma_resolve_addr");
 		ofi_straddr_log(&vrb_prov, FI_LOG_WARN, FI_LOG_EP_CTRL,
 				"src addr", rai->ai_src_addr);
 		ofi_straddr_log(&vrb_prov, FI_LOG_WARN, FI_LOG_EP_CTRL,
@@ -405,7 +423,7 @@ fn:
 }
 
 #if ENABLE_DEBUG
-static int vrb_dbg_query_qp_attr(struct ibv_qp *qp)
+static void vrb_dbg_query_qp_attr(struct ibv_qp *qp)
 {
 	struct ibv_qp_init_attr attr = { 0 };
 	struct ibv_qp_attr qp_attr = { 0 };
@@ -414,9 +432,10 @@ static int vrb_dbg_query_qp_attr(struct ibv_qp *qp)
 	ret = ibv_query_qp(qp, &qp_attr, IBV_QP_TIMEOUT | IBV_QP_RETRY_CNT |
 			   IBV_QP_RNR_RETRY | IBV_QP_MIN_RNR_TIMER, &attr);
 	if (ret) {
-		VERBS_WARN(FI_LOG_EP_CTRL, "Unable to query QP\n");
-		return ret;
+		VRB_WARN_ERRNO(FI_LOG_EP_CTRL, "ibv_query_qp");
+		return;
 	}
+
 	FI_DBG(&vrb_prov, FI_LOG_EP_CTRL, "QP attributes: "
 	       "min_rnr_timer"	": %" PRIu8 ", "
 	       "timeout"	": %" PRIu8 ", "
@@ -424,22 +443,20 @@ static int vrb_dbg_query_qp_attr(struct ibv_qp *qp)
 	       "rnr_retry"	": %" PRIu8 "\n",
 	       qp_attr.min_rnr_timer, qp_attr.timeout, qp_attr.retry_cnt,
 	       qp_attr.rnr_retry);
-	return 0;
 }
 #else
-static int vrb_dbg_query_qp_attr(struct ibv_qp *qp)
+static void vrb_dbg_query_qp_attr(struct ibv_qp *qp)
 {
-	return 0;
 }
 #endif
 
-int vrb_set_rnr_timer(struct ibv_qp *qp)
+void vrb_set_rnr_timer(struct ibv_qp *qp)
 {
 	struct ibv_qp_attr attr = { 0 };
 	int ret;
 
 	if (vrb_gl_data.min_rnr_timer > 31) {
-		VERBS_WARN(FI_LOG_EQ, "min_rnr_timer value out of valid range; "
+		VRB_WARN(FI_LOG_EQ, "min_rnr_timer value out of valid range; "
 			   "using default value of %d\n",
 			   VERBS_DEFAULT_MIN_RNR_TIMER);
 		attr.min_rnr_timer = VERBS_DEFAULT_MIN_RNR_TIMER;
@@ -449,17 +466,13 @@ int vrb_set_rnr_timer(struct ibv_qp *qp)
 
 	/* XRC initiator QP do not have responder logic */
 	if (qp->qp_type == IBV_QPT_XRC_SEND)
-		return 0;
+		return;
 
 	ret = ibv_modify_qp(qp, &attr, IBV_QP_MIN_RNR_TIMER);
-	if (ret) {
-		VERBS_WARN(FI_LOG_EQ, "Unable to modify QP attribute\n");
-		return ret;
-	}
-	ret = vrb_dbg_query_qp_attr(qp);
 	if (ret)
-		return ret;
-	return 0;
+		VRB_WARN_ERRNO(FI_LOG_EP_CTRL, "ibv_modify_qp");
+
+	vrb_dbg_query_qp_attr(qp);
 }
 
 int vrb_find_max_inline(struct ibv_pd *pd, struct ibv_context *context,
@@ -610,106 +623,104 @@ static int vrb_read_params(void)
 {
 	/* Common parameters */
 	if (vrb_get_param_int("tx_size", "Default maximum tx context size",
-				 &vrb_gl_data.def_tx_size) ||
+			      &vrb_gl_data.def_tx_size) ||
 	    (vrb_gl_data.def_tx_size < 0)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of tx_size\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of tx_size\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("rx_size", "Default maximum rx context size",
-				 &vrb_gl_data.def_rx_size) ||
+			      &vrb_gl_data.def_rx_size) ||
 	    (vrb_gl_data.def_rx_size < 0)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of rx_size\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of rx_size\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("tx_iov_limit", "Default maximum tx iov_limit",
-				 &vrb_gl_data.def_tx_iov_limit) ||
+			      &vrb_gl_data.def_tx_iov_limit) ||
 	    (vrb_gl_data.def_tx_iov_limit < 0)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of tx_iov_limit\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of tx_iov_limit\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("rx_iov_limit", "Default maximum rx iov_limit",
-				 &vrb_gl_data.def_rx_iov_limit) ||
+			      &vrb_gl_data.def_rx_iov_limit) ||
 	    (vrb_gl_data.def_rx_iov_limit < 0)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of rx_iov_limit\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of rx_iov_limit\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("inline_size", "Default maximum inline size. "
-				 "Actual inject size returned in fi_info may be "
-				 "greater", &vrb_gl_data.def_inline_size) ||
+			      "Actual inject size returned in fi_info may be "
+			      "greater", &vrb_gl_data.def_inline_size) ||
 	    (vrb_gl_data.def_inline_size < 0)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of inline_size\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of inline_size\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("min_rnr_timer", "Set min_rnr_timer QP "
-				 "attribute (0 - 31)",
-				 &vrb_gl_data.min_rnr_timer) ||
+			      "attribute (0 - 31)",
+			      &vrb_gl_data.min_rnr_timer) ||
 	    ((vrb_gl_data.min_rnr_timer < 0) ||
 	     (vrb_gl_data.min_rnr_timer > 31))) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of min_rnr_timer\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of min_rnr_timer\n");
 		return -FI_EINVAL;
 	}
 
 	if (vrb_get_param_bool("use_odp", "Enable on-demand paging memory "
-	    "registrations, if supported.  This is currently required to "
-	    "register DAX file system mmapped memory.", &vrb_gl_data.use_odp)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of use_odp\n");
+			       "registrations, if supported.  This is "
+			       "currently required to register DAX file system "
+			       "mmapped memory.", &vrb_gl_data.use_odp)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid value of use_odp\n");
 		return -FI_EINVAL;
 	}
 
-	if (vrb_get_param_bool("prefer_xrc", "Order XRC transport fi_infos"
-				  "ahead of RC. Default orders RC first.",
-				  &vrb_gl_data.msg.prefer_xrc)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of prefer_xrc\n");
+	if (vrb_get_param_bool("prefer_xrc", "Order XRC transport fi_infos "
+			       "ahead of RC.  Default orders RC first.  This "
+			       "setting must usually be combined with setting "
+			       "FI_OFI_RXM_USE_SRX.  See fi_verbs.7 man page.",
+				&vrb_gl_data.msg.prefer_xrc)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid value of prefer_xrc\n");
 		return -FI_EINVAL;
 	}
 
 	if (vrb_get_param_str("xrcd_filename", "A file to "
-				 "associate with the XRC domain.",
-				 &vrb_gl_data.msg.xrcd_filename)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of xrcd_filename\n");
+			      "associate with the XRC domain.",
+			      &vrb_gl_data.msg.xrcd_filename)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid value of xrcd_filename\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("cqread_bunch_size", "The number of entries to "
-				 "be read from the verbs completion queue at a time",
-				 &vrb_gl_data.cqread_bunch_size) ||
+			      "be read from the verbs completion queue at a time",
+			      &vrb_gl_data.cqread_bunch_size) ||
 	    (vrb_gl_data.cqread_bunch_size <= 0)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of cqread_bunch_size\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid value of cqread_bunch_size\n");
 		return -FI_EINVAL;
 	}
 	if (vrb_get_param_int("gid_idx", "Set which gid index to use "
-				 "attribute (0 - 255)",
-				 &vrb_gl_data.gid_idx) ||
-	    (vrb_gl_data.gid_idx < 0 ||
-	     vrb_gl_data.gid_idx > 255)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of gid index\n");
+			      "attribute (0 - 255)", &vrb_gl_data.gid_idx) ||
+	    (vrb_gl_data.gid_idx < 0 || vrb_gl_data.gid_idx > 255)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid value of gid index\n");
 		return -FI_EINVAL;
 	}
 
 	if (vrb_get_param_str("device_name", "The prefix or the full name of the "
-			      "verbs device to use",
-			      &vrb_gl_data.device_name)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of device_name\n");
+			      "verbs device to use", &vrb_gl_data.device_name)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid value of device_name\n");
 		return -FI_EINVAL;
 	}
 
+	if (vrb_gl_data.dmabuf_support) {
+		if (vrb_get_param_bool("use_dmabuf", "Enable dmabuf based memory "
+				       "registrations, if supported. Yes by default.",
+				       (int *)&vrb_gl_data.dmabuf_support)) {
+			VRB_WARN(FI_LOG_CORE, "Invalid value of use_dmabuf\n");
+			return -FI_EINVAL;
+		}
+	}
+	VRB_INFO(FI_LOG_CORE, "dmabuf support is %s\n",
+		 vrb_gl_data.dmabuf_support ? "enabled" : "disabled");
+
 	/* MSG-specific parameter */
 	if (vrb_get_param_str("iface", "The prefix or the full name of the "
-				 "network interface associated with the verbs device",
-				 &vrb_gl_data.iface)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of iface\n");
+			      "network interface associated with the verbs "
+			      "device", &vrb_gl_data.iface)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid value of iface\n");
 		return -FI_EINVAL;
 	}
 
@@ -717,21 +728,19 @@ static int vrb_read_params(void)
 	if (getenv("OMPI_COMM_WORLD_RANK") || getenv("PMI_RANK"))
 		vrb_gl_data.dgram.use_name_server = 0;
 	if (vrb_get_param_bool("dgram_use_name_server", "The option that "
-				  "enables/disables OFI Name Server thread that is used "
-				  "to resolve IP-addresses to provider specific "
-				  "addresses. If MPI is used, the NS is disabled "
-				  "by default.", &vrb_gl_data.dgram.use_name_server)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of dgram_use_name_server\n");
+			       "enables/disables OFI Name Server thread used "
+			       "to resolve IP-addresses to provider specific "
+			       "addresses. If MPI is used, the NS is disabled "
+			       "by default.", &vrb_gl_data.dgram.use_name_server)) {
+		VRB_WARN(FI_LOG_CORE, "Invalid dgram_use_name_server\n");
 		return -FI_EINVAL;
 	}
-	if (vrb_get_param_int("dgram_name_server_port", "The port on which Name Server "
-				 "thread listens incoming connections and requestes.",
-				 &vrb_gl_data.dgram.name_server_port) ||
+	if (vrb_get_param_int("dgram_name_server_port", "The port on which "
+			      "the name server thread listens incoming "
+			      "requests.", &vrb_gl_data.dgram.name_server_port) ||
 	    (vrb_gl_data.dgram.name_server_port < 0 ||
 	     vrb_gl_data.dgram.name_server_port > 65535)) {
-		VERBS_WARN(FI_LOG_CORE,
-			   "Invalid value of dgram_name_server_port\n");
+		VRB_WARN(FI_LOG_CORE, "Invalid dgram_name_server_port\n");
 		return -FI_EINVAL;
 	}
 
@@ -755,6 +764,50 @@ static void verbs_devs_free(void)
 	}
 }
 
+static void vrb_set_peer_mem_support(void)
+{
+	char *line = NULL;
+	size_t line_size = 0;
+	ssize_t bytes;
+	FILE *kallsyms_fd;
+
+	kallsyms_fd = fopen("/proc/kallsyms", "r");
+	if (!kallsyms_fd)
+		return;
+
+	while ((bytes = getline(&line, &line_size, kallsyms_fd)) != -1) {
+		if (strstr(line, "ib_register_peer_memory_client")) {
+			vrb_gl_data.peer_mem_support = true;
+			break;
+		}
+	}
+
+	free(line);
+	fclose(kallsyms_fd);
+}
+
+static void vrb_set_dmabuf_support(void)
+{
+	char *line = NULL;
+	size_t line_size = 0;
+	ssize_t bytes;
+	FILE *kallsyms_fd;
+
+	kallsyms_fd = fopen("/proc/kallsyms", "r");
+	if (!kallsyms_fd)
+		return;
+
+	while ((bytes = getline(&line, &line_size, kallsyms_fd)) != -1) {
+		if (strstr(line, "ib_umem_dmabuf_get")) {
+			vrb_gl_data.dmabuf_support = true;
+			break;
+		}
+	}
+
+	free(line);
+	fclose(kallsyms_fd);
+}
+
 static void vrb_fini(void)
 {
 #if HAVE_VERBS_DL
@@ -764,6 +817,7 @@ static void vrb_fini(void)
 #endif
 	fi_freeinfo((void *)vrb_util_prov.info);
 	verbs_devs_free();
+	vrb_os_fini();
 	vrb_util_prov.info = NULL;
 }
 
@@ -774,7 +828,11 @@ VERBS_INI
 	ofi_hmem_init();
 	ofi_monitors_init();
 #endif
-	if (vrb_read_params()|| vrb_init_info(&vrb_util_prov.info))
+
+	vrb_os_mem_support();
+
+	if (vrb_read_params() || vrb_os_ini() ||
+	    vrb_init_info(&vrb_util_prov.info))
 		return NULL;
 	return &vrb_prov;
 }
