@@ -1,5 +1,6 @@
 /*
  * Copyright (c) 2013-2017 Intel Corporation. All rights reserved.
+ * Copyright (c) 2022 DataDirect Networks, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -30,6 +31,8 @@
  * SOFTWARE.
  */
 
+#include "config.h"
+
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -503,11 +506,11 @@ void freeifaddrs(struct ifaddrs *ifa)
 	}
 }
 
-static ssize_t
+ssize_t
 ofi_sendv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt, int flags)
 {
-	ssize_t size = 0;
-	int ret, i;
+	ssize_t size = 0, ret;
+	int i;
 
 	if (iov_cnt == 1) {
 		return ofi_send_socket(fd, iovec[0].iov_base,
@@ -519,7 +522,7 @@ ofi_sendv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt, int flags
 				      iovec[i].iov_len, flags);
 		if (ret >= 0) {
 			size += ret;
-			if (ret != iovec[i].iov_len)
+			if ((size_t) ret != iovec[i].iov_len)
 				return size;
 		} else {
 			return size ? size : ret;
@@ -528,11 +531,11 @@ ofi_sendv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt, int flags
 	return size;
 }
 
-static ssize_t
+ssize_t
 ofi_recvv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt, int flags)
 {
-	ssize_t size = 0;
-	int ret, i;
+	ssize_t size = 0, ret;
+	int i;
 
 	if (iov_cnt == 1) {
 		return ofi_recv_socket(fd, iovec[0].iov_base,
@@ -544,7 +547,7 @@ ofi_recvv_socket(SOCKET fd, const struct iovec *iovec, size_t iov_cnt, int flags
 				      iovec[i].iov_len, flags);
 		if (ret >= 0) {
 			size += ret;
-			if (ret != iovec[i].iov_len)
+			if ((size_t) ret != iovec[i].iov_len)
 				return size;
 		} else {
 			return size ? size : ret;
@@ -597,6 +600,50 @@ ssize_t ofi_recvmsg_udp(SOCKET fd, struct msghdr *msg, int flags)
 			return ret;
 	}
 
-	ret = WSARecvMsg(fd, msg, &bytes, NULL, NULL);
+	ret = WSARecvMsg(fd, (LPWSAMSG) msg, &bytes, NULL, NULL);
 	return ret ? ret : bytes;
 }
+
+
+struct ofi_pollfds_ctx *ofi_pollfds_get_ctx(struct ofi_pollfds *pfds, int fd)
+{
+	struct ofi_pollfds_ctx *ctx = NULL;
+	int i;
+
+	assert(ofi_genlock_held(&pfds->lock));
+
+	/* 0 is signaling fd */
+	for (i = 1; i < pfds->size; i++) {
+		ctx = &pfds->ctx[i];
+		if (ctx->index >= 0 && ctx->index < pfds->nfds &&
+		    pfds->fds[ctx->index].fd == fd)
+			return ctx;
+	}
+
+	return NULL;
+}
+
+struct ofi_pollfds_ctx *ofi_pollfds_alloc_ctx(struct ofi_pollfds *pfds, int fd)
+{
+	struct ofi_pollfds_ctx *ctx;
+	int i;
+
+	assert(ofi_genlock_held(&pfds->lock));
+	assert(!ofi_pollfds_get_ctx(pfds, fd));
+	/* 0 is signaling fd */
+	for (i = 1; i < pfds->size; i++) {
+		if (pfds->ctx[i].index == -1) {
+			ctx = &pfds->ctx[i];
+			goto insert;
+		}
+	}
+
+	if (ofi_pollfds_grow(pfds, pfds->size + 1))
+		return NULL;
+
+	ctx = &pfds->ctx[pfds->nfds];
+insert:
+	assert(ctx->index < 0);
+	ctx->index = pfds->nfds++;
+	return ctx;
+}
